"""SSE event models."""

from __future__ import annotations

from typing import Any, Literal, Self

from pydantic import Field

from agentpool_server.opencode_server.models.app import Project  # noqa: TC001
from agentpool_server.opencode_server.models.base import OpenCodeBaseModel
from agentpool_server.opencode_server.models.common import FileDiff  # noqa: TC001
from agentpool_server.opencode_server.models.message import MessageInfo  # noqa: TC001
from agentpool_server.opencode_server.models.parts import Part  # noqa: TC001
from agentpool_server.opencode_server.models.pty import PtyInfo  # noqa: TC001
from agentpool_server.opencode_server.models.question import (  # noqa: TC001
    QuestionInfo,
    QuestionToolInfo,
)
from agentpool_server.opencode_server.models.session import (  # noqa: TC001
    Session,
    SessionStatus,
    SessionStatusType,
)


Variant = Literal["info", "success", "warning", "error"]
TodoPriority = Literal["high", "medium", "low"]
FileUpdateEvent = Literal["add", "change", "unlink"]
ConnectionStatus = Literal["connected", "error"]


class EmptyProperties(OpenCodeBaseModel):
    """Empty properties object."""


class ServerHeartbeatEvent(OpenCodeBaseModel):
    """Server heartbeat event - sent periodically to keep connections alive."""

    type: Literal["server.heartbeat"] = Field(default="server.heartbeat", init=False)
    properties: EmptyProperties = Field(default_factory=EmptyProperties)


class ServerConnectedEvent(OpenCodeBaseModel):
    """Server connected event."""

    type: Literal["server.connected"] = Field(default="server.connected", init=False)
    properties: EmptyProperties = Field(default_factory=EmptyProperties)


class SessionInfoProperties(OpenCodeBaseModel):
    """Session info wrapper for events."""

    info: Session


class SessionCreatedEvent(OpenCodeBaseModel):
    """Session created event."""

    type: Literal["session.created"] = Field(default="session.created", init=False)
    properties: SessionInfoProperties

    @classmethod
    def create(cls, session: Session) -> Self:
        return cls(properties=SessionInfoProperties(info=session))


class SessionUpdatedEvent(OpenCodeBaseModel):
    """Session updated event."""

    type: Literal["session.updated"] = Field(default="session.updated", init=False)
    properties: SessionInfoProperties

    @classmethod
    def create(cls, session: Session) -> Self:
        return cls(properties=SessionInfoProperties(info=session))


class SessionDeletedProperties(OpenCodeBaseModel):
    """Properties for session deleted event."""

    session_id: str


class SessionDeletedEvent(OpenCodeBaseModel):
    """Session deleted event."""

    type: Literal["session.deleted"] = Field(default="session.deleted", init=False)
    properties: SessionDeletedProperties

    @classmethod
    def create(cls, session_id: str) -> Self:
        return cls(properties=SessionDeletedProperties(session_id=session_id))


class SessionStatusProperties(OpenCodeBaseModel):
    """Properties for session status event."""

    session_id: str
    status: SessionStatus


class SessionStatusEvent(OpenCodeBaseModel):
    """Session status event."""

    type: Literal["session.status"] = Field(default="session.status", init=False)
    properties: SessionStatusProperties

    @classmethod
    def create(cls, session_id: str, status_type: SessionStatusType | SessionStatus) -> Self:
        status = SessionStatus(type=status_type) if isinstance(status_type, str) else status_type
        return cls(properties=SessionStatusProperties(session_id=session_id, status=status))


class SessionIdleProperties(OpenCodeBaseModel):
    """Properties for session idle event (deprecated but still used by TUI)."""

    session_id: str


class SessionIdleEvent(OpenCodeBaseModel):
    """Session idle event (deprecated but still used by TUI run command)."""

    type: Literal["session.idle"] = Field(default="session.idle", init=False)
    properties: SessionIdleProperties

    @classmethod
    def create(cls, session_id: str) -> Self:
        return cls(properties=SessionIdleProperties(session_id=session_id))


class SessionCompactedProperties(OpenCodeBaseModel):
    """Properties for session compacted event."""

    session_id: str


class SessionCompactedEvent(OpenCodeBaseModel):
    """Session compacted event - emitted when context compaction completes."""

    type: Literal["session.compacted"] = Field(default="session.compacted", init=False)
    properties: SessionCompactedProperties

    @classmethod
    def create(cls, session_id: str) -> Self:
        return cls(properties=SessionCompactedProperties(session_id=session_id))


class SessionErrorInfo(OpenCodeBaseModel):
    """Error information for session error event.

    Simplified version of OpenCode's error types (ProviderAuthError, UnknownError, etc.)
    """

    name: str
    """Error type name (e.g., 'UnknownError', 'ProviderAuthError')."""

    data: dict[str, Any] | None = None
    """Additional error data, typically contains 'message' field."""


class SessionErrorProperties(OpenCodeBaseModel):
    """Properties for session error event."""

    session_id: str | None = Field(default=None)
    error: SessionErrorInfo | None = None


class SessionErrorEvent(OpenCodeBaseModel):
    """Session error event - emitted when an error occurs during message processing."""

    type: Literal["session.error"] = Field(default="session.error", init=False)
    properties: SessionErrorProperties

    @classmethod
    def from_exception(cls, exception: Exception, session_id: str | None = None) -> Self:
        error_name = type(exception).__name__
        error_message = str(exception)
        return cls.create(session_id=session_id, error_name=error_name, error_message=error_message)

    @classmethod
    def create(
        cls,
        session_id: str | None = None,
        error_name: str = "UnknownError",
        error_message: str | None = None,
    ) -> Self:
        error_data = {"message": error_message} if error_message else None
        error = SessionErrorInfo(name=error_name, data=error_data)
        props = SessionErrorProperties(session_id=session_id, error=error)
        return cls(properties=props)


class MessageUpdatedEventProperties(OpenCodeBaseModel):
    """Properties for message updated event."""

    info: MessageInfo


class MessageUpdatedEvent(OpenCodeBaseModel):
    """Message updated event."""

    type: Literal["message.updated"] = Field(default="message.updated", init=False)
    properties: MessageUpdatedEventProperties

    @classmethod
    def create(cls, message: MessageInfo) -> Self:
        return cls(properties=MessageUpdatedEventProperties(info=message))


class PartUpdatedEventProperties(OpenCodeBaseModel):
    """Properties for part updated event."""

    part: Part
    delta: str | None = None


class PartUpdatedEvent(OpenCodeBaseModel):
    """Part updated event."""

    type: Literal["message.part.updated"] = Field(default="message.part.updated", init=False)
    properties: PartUpdatedEventProperties

    @classmethod
    def create(cls, part: Part, delta: str | None = None) -> Self:
        return cls(properties=PartUpdatedEventProperties(part=part, delta=delta))


class PartDeltaEventProperties(OpenCodeBaseModel):
    """Properties for message part delta event."""

    session_id: str
    message_id: str
    part_id: str
    delta: str


class PartDeltaEvent(OpenCodeBaseModel):
    """Message part delta event - streaming text delta for a part."""

    type: Literal["message.part.delta"] = Field(default="message.part.delta", init=False)
    properties: PartDeltaEventProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        message_id: str,
        part_id: str,
        delta: str,
    ) -> Self:
        return cls(
            properties=PartDeltaEventProperties(
                session_id=session_id,
                message_id=message_id,
                part_id=part_id,
                delta=delta,
            )
        )


class MessageRemovedProperties(OpenCodeBaseModel):
    """Properties for message removed event."""

    session_id: str
    message_id: str


class MessageRemovedEvent(OpenCodeBaseModel):
    """Message removed event - emitted during revert."""

    type: Literal["message.removed"] = Field(default="message.removed", init=False)
    properties: MessageRemovedProperties

    @classmethod
    def create(cls, session_id: str, message_id: str) -> Self:
        """Create message removed event."""
        props = MessageRemovedProperties(session_id=session_id, message_id=message_id)
        return cls(properties=props)


class PartRemovedProperties(OpenCodeBaseModel):
    """Properties for part removed event."""

    session_id: str
    message_id: str
    part_id: str


class PartRemovedEvent(OpenCodeBaseModel):
    """Part removed event - emitted during revert."""

    type: Literal["message.part.removed"] = Field(default="message.part.removed", init=False)
    properties: PartRemovedProperties

    @classmethod
    def create(cls, session_id: str, message_id: str, part_id: str) -> Self:
        """Create part removed event."""
        props = PartRemovedProperties(session_id=session_id, message_id=message_id, part_id=part_id)
        return cls(properties=props)


PermissionReply = Literal["once", "always", "reject"]
"""Permission reply type matching OpenCode's PermissionNext.Reply."""


class PermissionReplyRequest(OpenCodeBaseModel):
    """Request body for responding to a permission request."""

    reply: PermissionReply
    """Reply: 'once' | 'always' | 'reject'."""

    message: str | None = None
    """Optional message to include with the reply."""


class PermissionToolInfo(OpenCodeBaseModel):
    """Tool information for permission event."""

    message_id: str
    """Message ID."""

    call_id: str | None = None
    """Optional tool call ID."""


class PermissionAskedProperties(OpenCodeBaseModel):
    """Properties for permission.asked event.

    Matches OpenCode's PermissionNext.Event.Asked schema.
    """

    id: str
    """Permission request ID."""

    session_id: str
    """Session ID."""

    permission: str
    """Tool/permission type name."""

    patterns: list[str]
    """Patterns for matching (e.g., file paths, commands)."""

    metadata: dict[str, Any]
    """Arbitrary metadata about the tool call."""

    always: list[str]
    """Patterns that would be approved for future requests if user selects 'always'."""

    tool: PermissionToolInfo
    """Tool call information."""


class PermissionRequestEvent(OpenCodeBaseModel):
    """Permission request event - sent when a tool needs user confirmation.

    Uses 'permission.asked' event type for OpenCode TUI compatibility.
    """

    type: Literal["permission.asked"] = Field(default="permission.asked", init=False)
    properties: PermissionAskedProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        permission_id: str,
        tool_name: str,
        args_preview: str,
        message: str,
        message_id: str = "",
        call_id: str | None = None,
    ) -> Self:
        # Create pattern from tool name and args
        pattern = f"{tool_name}: {args_preview}" if args_preview else tool_name

        props = PermissionAskedProperties(
            id=permission_id,
            session_id=session_id,
            permission=tool_name,
            patterns=[pattern],
            metadata={"args_preview": args_preview},
            always=[pattern],  # Same pattern for "always" approval
            tool=PermissionToolInfo(message_id=message_id, call_id=call_id),
        )
        return cls(properties=props)


class PermissionRepliedProperties(OpenCodeBaseModel):
    """Properties for permission replied event.

    Matches OpenCode's permission.replied event schema.
    """

    session_id: str
    """Session ID."""

    request_id: str
    """Request/Permission ID."""

    reply: PermissionReply
    """Reply: 'once' | 'always' | 'reject'."""


class PermissionResolvedEvent(OpenCodeBaseModel):
    """Permission resolved event - sent when a permission request is answered.

    Uses 'permission.replied' event type for OpenCode TUI compatibility.
    """

    type: Literal["permission.replied"] = Field(default="permission.replied", init=False)
    properties: PermissionRepliedProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        request_id: str,
        reply: PermissionReply,
    ) -> Self:
        props = PermissionRepliedProperties(
            session_id=session_id,
            request_id=request_id,
            reply=reply,
        )
        return cls(properties=props)


class PermissionUpdatedProperties(OpenCodeBaseModel):
    """Properties for permission updated event."""

    id: str
    """Permission request ID."""

    session_id: str
    """Session ID."""

    permission: str
    """Tool/permission type name."""

    patterns: list[str]
    """Patterns for matching."""

    metadata: dict[str, Any]
    """Arbitrary metadata about the tool call."""

    always: list[str]
    """Patterns for 'always' approval."""

    tool: PermissionToolInfo
    """Tool call information."""


class PermissionUpdatedEvent(OpenCodeBaseModel):
    """Permission updated event - sent when permission status changes."""

    type: Literal["permission.updated"] = Field(default="permission.updated", init=False)
    properties: PermissionUpdatedProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        permission_id: str,
        tool_name: str,
        patterns: list[str],
        metadata: dict[str, Any],
        message_id: str = "",
        call_id: str | None = None,
    ) -> Self:
        props = PermissionUpdatedProperties(
            id=permission_id,
            session_id=session_id,
            permission=tool_name,
            patterns=patterns,
            metadata=metadata,
            always=patterns,
            tool=PermissionToolInfo(message_id=message_id, call_id=call_id),
        )
        return cls(properties=props)


# =============================================================================
# TUI Events - for external control of the TUI (e.g., VSCode extension)
# =============================================================================


class TuiPromptAppendProperties(OpenCodeBaseModel):
    """Properties for TUI prompt append event."""

    text: str


class TuiPromptAppendEvent(OpenCodeBaseModel):
    """TUI prompt append event - appends text to the prompt input."""

    type: Literal["tui.prompt.append"] = Field(default="tui.prompt.append", init=False)
    properties: TuiPromptAppendProperties

    @classmethod
    def create(cls, text: str) -> Self:
        return cls(properties=TuiPromptAppendProperties(text=text))


class TuiCommandExecuteProperties(OpenCodeBaseModel):
    """Properties for TUI command execute event."""

    command: str


class TuiCommandExecuteEvent(OpenCodeBaseModel):
    """TUI command execute event - executes a TUI command.

    Commands include:
    - session.list, session.new, session.share, session.interrupt, session.compact
    - session.page.up, session.page.down, session.half.page.up, session.half.page.down
    - session.first, session.last
    - prompt.clear, prompt.submit
    - agent.cycle
    """

    type: Literal["tui.command.execute"] = Field(default="tui.command.execute", init=False)
    properties: TuiCommandExecuteProperties

    @classmethod
    def create(cls, command: str) -> Self:
        return cls(properties=TuiCommandExecuteProperties(command=command))


class TuiToastShowProperties(OpenCodeBaseModel):
    """Properties for TUI toast show event."""

    title: str | None = None
    message: str
    variant: Variant = "info"
    duration: int = 5000  # Duration in milliseconds


class TuiToastShowEvent(OpenCodeBaseModel):
    """TUI toast show event - shows a toast notification."""

    type: Literal["tui.toast.show"] = Field(default="tui.toast.show", init=False)
    properties: TuiToastShowProperties

    @classmethod
    def create(
        cls,
        message: str,
        variant: Variant = "info",
        title: str | None = None,
        duration: int = 5000,
    ) -> Self:
        props = TuiToastShowProperties(
            title=title,
            message=message,
            variant=variant,
            duration=duration,
        )
        return cls(properties=props)


class TuiSessionSelectProperties(OpenCodeBaseModel):
    """Properties for TUI session select event."""

    session_id: str


class TuiSessionSelectEvent(OpenCodeBaseModel):
    """TUI session select event - navigates TUI to a specific session."""

    type: Literal["tui.session.select"] = Field(default="tui.session.select", init=False)
    properties: TuiSessionSelectProperties

    @classmethod
    def create(cls, session_id: str) -> Self:
        return cls(properties=TuiSessionSelectProperties(session_id=session_id))


# =============================================================================
# Todo Events
# =============================================================================


class Todo(OpenCodeBaseModel):
    """A single todo item."""

    id: str
    """Unique identifier for the todo item."""

    content: str
    """Brief description of the task."""

    status: Literal["pending", "in_progress", "completed", "cancelled"]
    """Current status: pending, in_progress, completed, cancelled."""

    priority: TodoPriority
    """Priority level: high, medium, low."""


class TodoUpdatedProperties(OpenCodeBaseModel):
    """Properties for todo updated event."""

    session_id: str
    todos: list[Todo]


class TodoUpdatedEvent(OpenCodeBaseModel):
    """Todo list updated event."""

    type: Literal["todo.updated"] = Field(default="todo.updated", init=False)
    properties: TodoUpdatedProperties

    @classmethod
    def create(cls, session_id: str, todos: list[Todo]) -> Self:
        return cls(properties=TodoUpdatedProperties(session_id=session_id, todos=todos))


# =============================================================================
# File Watcher Events
# =============================================================================


class FileWatcherUpdatedProperties(OpenCodeBaseModel):
    """Properties for file watcher updated event."""

    file: str
    """Absolute path to the file that changed."""

    event: FileUpdateEvent
    """Type of change: add (created), change (modified), unlink (deleted)."""


class FileWatcherUpdatedEvent(OpenCodeBaseModel):
    """File watcher updated event - sent when a project file changes."""

    type: Literal["file.watcher.updated"] = Field(default="file.watcher.updated", init=False)
    properties: FileWatcherUpdatedProperties

    @classmethod
    def create(cls, file: str, event: FileUpdateEvent) -> Self:
        return cls(properties=FileWatcherUpdatedProperties(file=file, event=event))


# =============================================================================
# PTY Events
# =============================================================================


class PtyCreatedProperties(OpenCodeBaseModel):
    """Properties for PTY created event."""

    info: PtyInfo
    """PTY session info."""


class PtyCreatedEvent(OpenCodeBaseModel):
    """PTY session created event."""

    type: Literal["pty.created"] = Field(default="pty.created", init=False)
    properties: PtyCreatedProperties

    @classmethod
    def create(cls, info: PtyInfo) -> Self:
        return cls(properties=PtyCreatedProperties(info=info))


class PtyUpdatedProperties(OpenCodeBaseModel):
    """Properties for PTY updated event."""

    info: PtyInfo
    """PTY session info."""


class PtyUpdatedEvent(OpenCodeBaseModel):
    """PTY session updated event."""

    type: Literal["pty.updated"] = Field(default="pty.updated", init=False)
    properties: PtyUpdatedProperties

    @classmethod
    def create(cls, info: PtyInfo) -> Self:
        return cls(properties=PtyUpdatedProperties(info=info))


class PtyExitedProperties(OpenCodeBaseModel):
    """Properties for PTY exited event."""

    id: str
    """PTY session ID."""

    exit_code: int
    """Process exit code."""


class PtyExitedEvent(OpenCodeBaseModel):
    """PTY process exited event."""

    type: Literal["pty.exited"] = Field(default="pty.exited", init=False)
    properties: PtyExitedProperties

    @classmethod
    def create(cls, pty_id: str, exit_code: int) -> Self:
        return cls(properties=PtyExitedProperties(id=pty_id, exit_code=exit_code))


class PtyDeletedProperties(OpenCodeBaseModel):
    """Properties for PTY deleted event."""

    id: str
    """PTY session ID."""


class PtyDeletedEvent(OpenCodeBaseModel):
    """PTY session deleted event."""

    type: Literal["pty.deleted"] = Field(default="pty.deleted", init=False)
    properties: PtyDeletedProperties

    @classmethod
    def create(cls, pty_id: str) -> Self:
        return cls(properties=PtyDeletedProperties(id=pty_id))


# =============================================================================
# LSP Events
# =============================================================================


class LspStatus(OpenCodeBaseModel):
    """LSP server status information."""

    id: str
    """Server identifier (e.g., 'pyright', 'rust-analyzer')."""

    name: str
    """Server name."""

    root: str
    """Relative workspace root path."""

    status: ConnectionStatus
    """Connection status."""


class LspUpdatedEvent(OpenCodeBaseModel):
    """LSP status updated event - sent when LSP server status changes."""

    type: Literal["lsp.updated"] = Field(default="lsp.updated", init=False)
    properties: EmptyProperties = Field(default_factory=EmptyProperties)


class LspClientDiagnosticsProperties(OpenCodeBaseModel):
    """Properties for LSP client diagnostics event."""

    server_id: str
    """LSP server ID that produced the diagnostics."""

    path: str
    """File path the diagnostics apply to."""


class LspClientDiagnosticsEvent(OpenCodeBaseModel):
    """LSP client diagnostics event - sent when diagnostics are published."""

    type: Literal["lsp.client.diagnostics"] = Field(default="lsp.client.diagnostics", init=False)
    properties: LspClientDiagnosticsProperties

    @classmethod
    def create(cls, server_id: str, path: str) -> Self:
        return cls(properties=LspClientDiagnosticsProperties(server_id=server_id, path=path))


# =============================================================================
# VCS Events
# =============================================================================


class ProjectUpdatedEvent(OpenCodeBaseModel):
    """Project metadata updated event."""

    type: Literal["project.updated"] = Field(default="project.updated", init=False)
    properties: Project

    @classmethod
    def create(cls, project: Project) -> Self:
        """Create project updated event."""
        return cls(properties=project)


class VcsBranchUpdatedProperties(OpenCodeBaseModel):
    """Properties for VCS branch updated event."""

    branch: str | None = None
    """Current branch name, or None if detached HEAD."""


# =============================================================================
# Session Diff Events
# =============================================================================


class SessionDiffProperties(OpenCodeBaseModel):
    """Properties for session diff event."""

    session_id: str
    diff: list[FileDiff]


class SessionDiffEvent(OpenCodeBaseModel):
    """Session diff event - emitted when file diffs are computed (revert, summary)."""

    type: Literal["session.diff"] = Field(default="session.diff", init=False)
    properties: SessionDiffProperties

    @classmethod
    def create(cls, session_id: str, diff: list[FileDiff]) -> Self:
        return cls(properties=SessionDiffProperties(session_id=session_id, diff=diff))


# =============================================================================
# File Events
# =============================================================================


class FileEditedProperties(OpenCodeBaseModel):
    """Properties for file edited event."""

    file: str
    """Absolute path to the edited file."""


class FileEditedEvent(OpenCodeBaseModel):
    """File edited event - emitted when a tool edits/writes/patches a file."""

    type: Literal["file.edited"] = Field(default="file.edited", init=False)
    properties: FileEditedProperties

    @classmethod
    def create(cls, file: str) -> Self:
        return cls(properties=FileEditedProperties(file=file))


# =============================================================================
# MCP Events
# =============================================================================


class McpToolsChangedProperties(OpenCodeBaseModel):
    """Properties for MCP tools changed event."""

    server: str
    """Name of the MCP server whose tools changed."""


class McpToolsChangedEvent(OpenCodeBaseModel):
    """MCP tools changed event - emitted when an MCP server's tool list changes.

    TODO: Hook into MCP SDK's ToolListChangedNotification to emit this event.
    OpenCode only emits this from the notification handler, not on connect/disconnect.
    """

    type: Literal["mcp.tools.changed"] = Field(default="mcp.tools.changed", init=False)
    properties: McpToolsChangedProperties

    @classmethod
    def create(cls, server: str) -> Self:
        return cls(properties=McpToolsChangedProperties(server=server))


# =============================================================================
# Command Events
# =============================================================================


class CommandExecutedProperties(OpenCodeBaseModel):
    """Properties for command executed event."""

    name: str
    """Command name."""

    session_id: str
    """Session ID."""

    arguments: str
    """Command arguments."""

    message_id: str
    """ID of the message that resulted from the command."""


class CommandExecutedEvent(OpenCodeBaseModel):
    """Command executed event - emitted after a slash command runs."""

    type: Literal["command.executed"] = Field(default="command.executed", init=False)
    properties: CommandExecutedProperties

    @classmethod
    def create(
        cls,
        name: str,
        session_id: str,
        arguments: str,
        message_id: str,
    ) -> Self:
        return cls(
            properties=CommandExecutedProperties(
                name=name,
                session_id=session_id,
                arguments=arguments,
                message_id=message_id,
            )
        )


class VcsBranchUpdatedEvent(OpenCodeBaseModel):
    """VCS branch updated event - sent when git branch changes."""

    type: Literal["vcs.branch.updated"] = Field(default="vcs.branch.updated", init=False)
    properties: VcsBranchUpdatedProperties

    @classmethod
    def create(cls, branch: str | None) -> Self:
        return cls(properties=VcsBranchUpdatedProperties(branch=branch))


class QuestionAskedProperties(OpenCodeBaseModel):
    """Properties for question asked event."""

    id: str
    session_id: str
    questions: list[QuestionInfo]
    tool: QuestionToolInfo | None = None


class QuestionAskedEvent(OpenCodeBaseModel):
    """Question asked event - sent when agent asks a question."""

    type: Literal["question.asked"] = Field(default="question.asked", init=False)
    properties: QuestionAskedProperties

    @classmethod
    def create(
        cls,
        request_id: str,
        session_id: str,
        questions: list[QuestionInfo],
        tool: QuestionToolInfo | None = None,
    ) -> Self:
        props = QuestionAskedProperties(
            id=request_id,
            session_id=session_id,
            questions=questions,
            tool=tool,
        )
        return cls(properties=props)


class QuestionRepliedProperties(OpenCodeBaseModel):
    """Properties for question replied event."""

    session_id: str
    request_id: str
    answers: list[list[str]]


class QuestionRepliedEvent(OpenCodeBaseModel):
    """Question replied event - sent when user answers a question."""

    type: Literal["question.replied"] = Field(default="question.replied", init=False)
    properties: QuestionRepliedProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        request_id: str,
        answers: list[list[str]],
    ) -> Self:
        props = QuestionRepliedProperties(
            session_id=session_id,
            request_id=request_id,
            answers=answers,
        )
        return cls(properties=props)


class QuestionRejectedProperties(OpenCodeBaseModel):
    """Properties for question rejected event."""

    session_id: str
    request_id: str


class QuestionRejectedEvent(OpenCodeBaseModel):
    """Question rejected event - sent when user dismisses a question."""

    type: Literal["question.rejected"] = Field(default="question.rejected", init=False)
    properties: QuestionRejectedProperties

    @classmethod
    def create(
        cls,
        session_id: str,
        request_id: str,
    ) -> Self:
        props = QuestionRejectedProperties(session_id=session_id, request_id=request_id)
        return cls(properties=props)


Event = (
    ServerConnectedEvent
    | ServerHeartbeatEvent
    | SessionCreatedEvent
    | SessionUpdatedEvent
    | SessionDeletedEvent
    | SessionStatusEvent
    | SessionErrorEvent
    | SessionIdleEvent
    | SessionDiffEvent
    | SessionCompactedEvent
    | MessageUpdatedEvent
    | MessageRemovedEvent
    | PartUpdatedEvent
    | PartDeltaEvent
    | PartRemovedEvent
    | PermissionRequestEvent
    | PermissionResolvedEvent
    | PermissionUpdatedEvent
    | QuestionAskedEvent
    | QuestionRepliedEvent
    | QuestionRejectedEvent
    | TodoUpdatedEvent
    | FileWatcherUpdatedEvent
    | FileEditedEvent
    | McpToolsChangedEvent
    | CommandExecutedEvent
    | PtyCreatedEvent
    | PtyUpdatedEvent
    | PtyExitedEvent
    | PtyDeletedEvent
    | LspUpdatedEvent
    | LspClientDiagnosticsEvent
    | ProjectUpdatedEvent
    | VcsBranchUpdatedEvent
    | TuiPromptAppendEvent
    | TuiCommandExecuteEvent
    | TuiToastShowEvent
    | TuiSessionSelectEvent
)



"""Code execution provider with secure tool isolation via FastAPI server."""

from __future__ import annotations

import asyncio
import contextlib
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, Self

import anyenv
import anyio

from agentpool.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Sequence
    import socket
    from types import TracebackType

    from exxec.base import ExecutionEnvironment
    from exxec.models import ServerInfo
    from exxec_config import ExecutionEnvironmentConfig
    from fastapi import FastAPI
    from fsspec import AbstractFileSystem
    from schemez import ToolsetCodeGenerator
    import uvicorn

    from agentpool.tools.base import Tool


logger = get_logger(__name__)


@dataclass
class RemoteCodeExecutor:
    """Provides secure code execution with tool access via FastAPI server.

    Architecture:
    - FastAPI server runs in HOST environment with tool routes
    - User code runs in SANDBOX environment (Docker, E2B, etc.)
    - Sandbox makes HTTP calls to server for tool execution
    """

    toolset_generator: ToolsetCodeGenerator
    """Code generator for tools."""

    execution_env: ExecutionEnvironment
    """Execution environment for running code."""

    def __post_init__(self) -> None:
        """Initialize filesystem for script history after dataclass init."""
        # Fallback filesystem if none provided to execute_code
        from fsspec.implementations.memory import MemoryFileSystem

        self._fallback_fs = MemoryFileSystem()

    @classmethod
    def from_tools(
        cls,
        tools: Sequence[Tool],
        env_config: ExecutionEnvironmentConfig,
        server_host: str = "localhost",
        server_port: int = 8000,
        include_docstrings: bool = True,
    ) -> RemoteCodeExecutor:
        """Create provider from tools and environment configuration.

        Args:
            tools: Tools to make available for code execution
            env_config: Execution environment configuration
            server_host: Host for FastAPI server
            server_port: Port for FastAPI server
            include_docstrings: Include function docstrings in documentation

        Returns:
            RemoteCodeExecutor instance
        """
        from agentpool.resource_providers.codemode.helpers import tools_to_codegen

        toolset_gen = tools_to_codegen(tools, include_docstrings)
        server_handler = ToolServerLifecycleHandler(toolset_gen, server_host, server_port)
        execution_env = env_config.get_provider(server_handler)
        return cls(toolset_gen, execution_env)

    def get_tool_description(self) -> str:
        """Get comprehensive description of available tools."""
        return self.toolset_generator.generate_tool_description()

    async def execute_code(
        self,
        code: str,
        title: str,
        internal_fs: AbstractFileSystem | None = None,
    ) -> Any:
        """Execute code with tools available via HTTP API.

        Args:
            code: Python code to execute
            title: Short descriptive title for this script (3-4 words)
            internal_fs: Filesystem to write script history to (uses fallback if None)

        Returns:
            Execution result from the environment
        """
        from upathtools.async_ops import to_async_fs

        fs = internal_fs or self._fallback_fs
        fs = to_async_fs(fs)
        start_time = datetime.now(UTC)
        exit_code = 0
        error_msg: str | None = None
        exec_result = None
        result_str = ""

        try:
            exec_result = await self.execution_env.execute(code)
            result_str = str(exec_result)
            # Check if execution failed
            if exec_result.exit_code is not None and exec_result.exit_code != 0:
                exit_code = exec_result.exit_code
                error_msg = exec_result.error
        except Exception as e:  # noqa: BLE001
            exit_code = 1
            error_msg = str(e)
            result_str = f"Error executing code: {error_msg}"
        finally:
            # Save to filesystem (agent's internal_fs or fallback)
            end_time = datetime.now(UTC)
            duration = (end_time - start_time).total_seconds()
            timestamp = start_time.strftime("%Y%m%d_%H%M%S")

            # Write script file
            script_path = f"remote_code/scripts/{timestamp}_{title}.py"
            await fs._pipe_file(script_path, code.encode("utf-8"))

            # Write metadata file
            metadata = {
                "title": title,
                "timestamp": start_time.isoformat(),
                "exit_code": exit_code,
                "duration": duration,
                "result": result_str,
                "error": error_msg,
            }
            metadata_path = f"remote_code/scripts/{timestamp}_{title}.json"
            dumped = anyenv.dump_json(metadata, indent=True)
            await fs._pipe_file(metadata_path, dumped.encode("utf-8"))

        return exec_result

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        await self.execution_env.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async context manager exit."""
        return await self.execution_env.__aexit__(exc_type, exc_val, exc_tb)


class ToolServerLifecycleHandler:
    """Manages FastAPI server lifecycle for tool access."""

    def __init__(
        self,
        toolset_generator: ToolsetCodeGenerator,
        host: str = "localhost",
        port: int = 8000,
    ) -> None:
        self.toolset_generator = toolset_generator
        self.host = host
        self.port = port  # Will be set when socket is created
        self.app: FastAPI | None = None
        self.server: uvicorn.Server | None = None
        self._server_task: asyncio.Task[None] | None = None
        self._socket: socket.socket | None = None

    async def __aenter__(self) -> ServerInfo:
        """Start FastAPI server with tool routes."""
        from exxec.models import ServerInfo
        from fastapi import FastAPI
        import uvicorn

        from agentpool.utils.network import _create_socket

        if self.server is not None:
            return ServerInfo(url=f"http://{self.host}:{self.port}", port=self.port)
        # Create socket and get actual port
        self._socket, self.port = _create_socket(self.port)
        # Create FastAPI app
        self.app = FastAPI(title="Tool Server", description="Generated tool endpoints")
        # Add tool routes
        self.toolset_generator.add_all_routes(self.app, "/tools")
        config = uvicorn.Config(
            self.app,
            log_level="error",  # Reduce log noise
            access_log=False,
            host=self.host,
            port=self.port,
            ws="websockets-sansio",
        )
        self.server = uvicorn.Server(config)
        # Start server in background with our socket
        self._server_task = asyncio.create_task(self.server.serve([self._socket]))
        logger.info("Started tool server", host=self.host, port=self.port)
        await anyio.sleep(0.1)  # Wait for server to start properly
        # Verify server is running
        max_retries = 10
        for _ in range(max_retries):
            if self.server.started:
                break
            await anyio.sleep(0.1)

        return ServerInfo(url=f"http://{self.host}:{self.port}", port=self.port)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Stop FastAPI server."""
        # Stop server gracefully
        if self.server:
            self.server.should_exit = True
            self.server.force_exit = True

        # Cancel server task
        if self._server_task and not self._server_task.done():
            self._server_task.cancel()
            with contextlib.suppress(asyncio.CancelledError, Exception):
                await asyncio.wait_for(self._server_task, timeout=1.0)

        # Close socket
        if self._socket:
            with contextlib.suppress(Exception):
                self._socket.close()

        # Reset state
        self.server = None
        self._server_task = None
        self._socket = None
        self.app = None


if __name__ == "__main__":
    from exxec_config import LocalExecutionEnvironmentConfig

    from agentpool.tools.base import Tool

    def add_numbers(x: int, y: int) -> int:
        """Add two numbers."""
        return x + y

    async def main() -> None:
        tools = [Tool.from_callable(add_numbers)]
        config = LocalExecutionEnvironmentConfig()
        provider = RemoteCodeExecutor.from_tools(tools, config, server_port=9876)
        async with provider:
            result = await provider.execute_code(
                "_result = await add_numbers(5, 3)", title="test_addition"
            )
            print(f"Result: {result.result}")

    anyio.run(main)



"""Tool management commands."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from slashed import CommandContext, CommandError, CompletionContext  # noqa: TC002
from slashed.completers import CallbackCompleter

from agentpool.agents.context import AgentContext  # noqa: TC001
from agentpool.log import get_logger
from agentpool.utils.importing import import_callable
from agentpool_commands.base import NodeCommand
from agentpool_commands.markdown_utils import format_table


if TYPE_CHECKING:
    from agentpool.messaging import MessageNode


logger = get_logger(__name__)


class ListToolsCommand(NodeCommand):
    """Show all available tools and their current status.

    Tools are grouped by source (runtime/agent/builtin).
    ✓ indicates enabled, ✗ indicates disabled.
    """

    name = "list-tools"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(
        self,
        ctx: CommandContext[AgentContext],
        source: str | None = None,
    ) -> None:
        """List all available tools.

        Args:
            ctx: Command context
            source: Optional filter by source (runtime/agent/builtin)
        """
        agent = ctx.context.agent

        rows = [
            {
                "Status": "✅" if tool_info.enabled else "❌",
                "Name": tool_info.name,
                "Source": tool_info.source,
            }
            for tool_info in await agent.tools.get_tools()
            if not source or tool_info.source == source
        ]

        headers = ["Status", "Name", "Source"]
        table = format_table(headers, rows)
        await ctx.print(f"## 🔧 Available Tools\n\n{table}")


class ShowToolCommand(NodeCommand):
    """Display detailed information about a specific tool.

    Shows:
    - Source (runtime/agent/builtin)
    - Current status (enabled/disabled)
    - Parameter descriptions
    - Additional metadata

    Example: /tool-info open_browser
    """

    name = "show-tool"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(self, ctx: CommandContext[AgentContext], name: str) -> None:
        """Show detailed information about a tool.

        Args:
            ctx: Command context
            name: Tool name to show
        """
        agent = ctx.context.agent

        try:
            tool_info = await agent.tools.get_tool(name)
            # Start with the standard tool info format
            sections = [tool_info.format_info(indent="")]

            # Add extra metadata section if we have any additional info
            extra_info = []
            if tool_info.requires_confirmation:
                extra_info.append("Requires Confirmation: Yes")
            if tool_info.source != "dynamic":  # Only show if not default
                extra_info.append(f"Source: {tool_info.source}")
            if tool_info.metadata:
                extra_info.append("\nMetadata:")
                extra_info.extend(f"- {k}: {v}" for k, v in tool_info.metadata.items())

            if extra_info:
                sections.extend(extra_info)

            await ctx.print("\n".join(sections))
        except KeyError:
            await ctx.print(f"❌ **Tool** `{name}` **not found**")

    def get_completer(self) -> CallbackCompleter:
        """Get completer for tool names."""
        return CallbackCompleter(get_tool_names)


class EnableToolCommand(NodeCommand):
    """Enable a previously disabled tool.

    Use /list-tools to see available tools.

    Example: /enable-tool open_browser
    """

    name = "enable-tool"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(self, ctx: CommandContext[AgentContext], name: str) -> None:
        """Enable a tool.

        Args:
            ctx: Command context
            name: Tool name to enable
        """
        try:
            await ctx.context.agent.tools.enable_tool(name)
            await ctx.print(f"✅ **Tool** `{name}` **enabled**")
        except ValueError as e:
            raise CommandError(f"Failed to enable tool: {e}") from e

    def get_completer(self) -> CallbackCompleter:
        """Get completer for tool names."""
        return CallbackCompleter(get_tool_names)


class DisableToolCommand(NodeCommand):
    """Disable a tool to prevent its use.

    Use /list-tools to see available tools.

    Example: /disable-tool open_browser
    """

    name = "disable-tool"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(self, ctx: CommandContext[AgentContext], name: str) -> None:
        """Disable a tool.

        Args:
            ctx: Command context
            name: Tool name to disable
        """
        try:
            await ctx.context.agent.tools.disable_tool(name)
            await ctx.print(f"❌ **Tool** `{name}` **disabled**")
        except ValueError as e:
            raise CommandError(f"Failed to disable tool: {e}") from e

    def get_completer(self) -> CallbackCompleter:
        """Get completer for tool names."""
        return CallbackCompleter(get_tool_names)


class RegisterToolCommand(NodeCommand):
    """Register a new tool from a Python import path.

    Examples:
      /register-tool webbrowser.open
      /register-tool json.dumps --name format_json
      /register-tool os.getcwd --description 'Get current directory'
    """

    name = "register-tool"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(
        self,
        ctx: CommandContext[AgentContext],
        import_path: str,
        *,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        """Register a new tool from import path or function.

        Args:
            ctx: Command context
            import_path: Python import path to the function
            name: Optional custom name for the tool
            description: Optional custom description
        """
        try:
            callable_func = import_callable(import_path)
            # Register with ToolManager
            tool_info = ctx.context.agent.tools.register_tool(
                callable_func,
                name_override=name,
                description_override=description,
                enabled=True,
                source="dynamic",
                metadata={"import_path": import_path, "registered_via": "register-tool"},
            )

            # Show the registered tool info
            tool_info.format_info()
            await ctx.print(
                f"✅ **Tool registered successfully:**\n`{tool_info.name}`"
                f" - {tool_info.description or '*No description*'}"
            )

        except Exception as e:
            raise CommandError(f"Failed to register tool: {e}") from e


class RegisterCodeToolCommand(NodeCommand):
    """Register a new tool from Python code.

    The code should define a function that will become the tool.
    The function name becomes the tool name unless overridden.

    Options:
      --name tool_name         Custom name for the tool
      --description "text"     Custom description

    Examples:
      /register-code-tool "def greet(name: str) -> str: return f'Hello {name}'"
      /register-code-tool "def add(a: int, b: int) -> int: return a + b" --name sum
    """

    name = "register-code-tool"
    category = "tools"

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        from agentpool import Agent

        return isinstance(node, Agent)

    async def execute_command(
        self,
        ctx: CommandContext[AgentContext],
        code: str,
        *,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        """Register a new tool from code string.

        Args:
            ctx: Command context
            code: Python code defining a function
            name: Optional custom name for the tool
            description: Optional custom description
        """
        from agentpool.tools.base import Tool

        try:
            tool_obj = Tool.from_code(code, name=name, description=description)
            tool_obj.enabled = True
            tool_obj.source = "dynamic"

            registered = ctx.context.agent.tools.register_tool(tool_obj)
            await ctx.print(
                f"**Tool registered:** `{registered.name}`"
                f" - {registered.description or '*No description*'}"
            )
        except Exception as e:
            raise CommandError(f"Failed to register code tool: {e}") from e


async def get_tool_names(ctx: CompletionContext[AgentContext]) -> list[str]:
    manager = ctx.command_context.context.agent.tools
    return list({t.name for t in await manager.get_tools()})



"""Config and provider routes."""

from __future__ import annotations

from collections import defaultdict
from datetime import timedelta
import os
from typing import TYPE_CHECKING

from fastapi import APIRouter

from agentpool_server.opencode_server.dependencies import StateDep
from agentpool_server.opencode_server.models import (
    Config,
    Mode,
    Model,
    ModelCost,
    ModelLimit,
    Provider,
    ProviderListResponse,
    ProvidersResponse,
)


if TYPE_CHECKING:
    from tokonomics.model_discovery.model_info import ModelInfo as TokoModelInfo

    from agentpool.agents.base_agent import BaseAgent


router = APIRouter(tags=["config"])

# Provider display names and environment variable mappings
PROVIDER_INFO: dict[str, tuple[str, list[str]]] = {
    "anthropic": ("Anthropic", ["ANTHROPIC_API_KEY"]),
    "openai": ("OpenAI", ["OPENAI_API_KEY"]),
    "google": ("Google", ["GOOGLE_API_KEY", "GEMINI_API_KEY"]),
    "mistral": ("Mistral", ["MISTRAL_API_KEY"]),
    "groq": ("Groq", ["GROQ_API_KEY"]),
    "deepseek": ("DeepSeek", ["DEEPSEEK_API_KEY"]),
    "xai": ("xAI", ["XAI_API_KEY"]),
    "together": ("Together AI", ["TOGETHER_API_KEY"]),
    "perplexity": ("Perplexity", ["PERPLEXITY_API_KEY"]),
    "cohere": ("Cohere", ["COHERE_API_KEY"]),
    "fireworks": ("Fireworks AI", ["FIREWORKS_API_KEY"]),
    "openrouter": ("OpenRouter", ["OPENROUTER_API_KEY"]),
    "bedrock": ("AWS Bedrock", ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]),
    "azure": ("Azure OpenAI", ["AZURE_OPENAI_API_KEY", "AZURE_OPENAI_ENDPOINT"]),
    "vertex": ("Google Vertex AI", ["GOOGLE_APPLICATION_CREDENTIALS"]),
}


def _group_models_by_provider(models: list[TokoModelInfo]) -> dict[str, list[TokoModelInfo]]:
    """Group models by their provider."""
    grouped: dict[str, list[TokoModelInfo]] = defaultdict(list)
    for model in models:
        # Skip embedding models - OpenCode is for chat/agent models
        if model.is_embedding:
            continue
        grouped[model.provider].append(model)
    return grouped


def _build_providers(models: list[TokoModelInfo]) -> list[Provider]:
    """Build Provider list from tokonomics models."""
    grouped = _group_models_by_provider(models)
    providers: list[Provider] = []
    for provider_id, provider_models in sorted(grouped.items()):
        # Get provider display info
        display_name, env_vars = PROVIDER_INFO.get(
            provider_id, (provider_id.title(), [f"{provider_id.upper()}_API_KEY"])
        )
        # Convert models to OpenCode format
        models_dict = {i.id_override or i.id: Model.from_tokonomics(i) for i in provider_models}
        provider = Provider(id=provider_id, name=display_name, env=env_vars, models=models_dict)
        providers.append(provider)

    return providers


async def _get_model_providers(agent: BaseAgent) -> list[Provider]:
    providers: list[Provider] = []
    # Try to get models from the agent
    try:
        if toko_models := await agent.get_available_models():
            providers = _build_providers(toko_models)
    except Exception:  # noqa: BLE001
        pass  # Fall through to dummy providers

    # Fall back to dummy providers if no models available
    if not providers:
        providers = _get_dummy_providers()

    # Get variants from agent's thought_level modes (for Codex, Claude Code, etc.)
    if variants := await _get_variants_from_agent(agent):
        providers = _apply_variants_to_providers(providers, variants)
    return providers


async def _get_available_models() -> list[TokoModelInfo]:
    """Fetch available models using tokonomics."""
    from tokonomics.model_discovery import get_all_models

    max_age = timedelta(days=7)  # Cache for a week
    return await get_all_models(max_age=max_age)


async def _get_variants_from_agent(agent: object) -> dict[str, dict[str, object]]:
    """Get variants from agent's thought_level modes.

    Only supported for Codex and Claude Code agents which have static,
    known thought_level modes.

    Args:
        agent: The agent to get modes from

    Returns:
        Dict mapping variant names to empty config dicts (config is agent-internal)
    """
    from agentpool.agents.claude_code_agent import ClaudeCodeAgent
    from agentpool.agents.codex_agent import CodexAgent

    # Only Codex and Claude Code have static thought_level modes we can expose
    if not isinstance(agent, (CodexAgent, ClaudeCodeAgent)):
        return {}

    try:
        mode_categories = await agent.get_modes()
    except Exception:  # noqa: BLE001
        return {}
    for category in mode_categories:
        if category.id == "thought_level":
            # Convert modes to variants - the actual config is handled by set_mode
            return {mode.id: {} for mode in category.available_modes}
    return {}


def _apply_variants_to_providers(
    providers: list[Provider], variants: dict[str, dict[str, object]]
) -> list[Provider]:
    """Apply variants to all models in all providers.

    For agents with known thought_level modes (Codex, Claude Code),
    the same variants apply to all models.
    """
    if not variants:
        return providers

    updated_providers = []
    for provider in providers:
        updated_models = {
            model_id: model.model_copy(update={"variants": variants})
            for model_id, model in provider.models.items()
        }
        updated_providers.append(provider.model_copy(update={"models": updated_models}))
    return updated_providers


def _get_dummy_providers() -> list[Provider]:
    """Return a single dummy provider for testing."""
    dummy_model = Model(
        id="gpt-4o",
        name="GPT-4o",
        attachment=True,
        cost=ModelCost(input=5.0, output=15.0),
        limit=ModelLimit(context=128000.0, output=4096.0),
        reasoning=False,
        release_date="2024-05-13",
        temperature=True,
        tool_call=True,
    )
    dummy_provider = Provider(
        id="openai",
        name="OpenAI",
        env=["OPENAI_API_KEY"],
        models={"gpt-4o": dummy_model},
    )
    return [dummy_provider]


@router.get("/config")
async def get_config(state: StateDep) -> Config:
    """Get server configuration."""
    # Set a default model if not already configured
    if state.config.model is None:
        try:
            if toko_models := await state.agent.get_available_models():
                providers = _build_providers(toko_models)
                # Find first connected provider and use its first model
                for provider in providers:
                    if any(os.environ.get(env) for env in provider.env) and provider.models:
                        first_model = next(iter(provider.models.keys()))
                        state.config.model = f"{provider.id}/{first_model}"
                        break
        except Exception:  # noqa: BLE001
            pass  # If we can't set a default, that's okay

    return state.config


@router.patch("/config")
async def update_config(state: StateDep, config_update: Config) -> Config:
    """Update server configuration.

    Only updates fields that are provided (non-None).
    Returns the complete updated config.
    """
    # Update only the fields that were provided
    update_data = config_update.model_dump(exclude_unset=True)
    for field_name, value in update_data.items():
        setattr(state.config, field_name, value)

    return state.config


@router.get("/config/providers")
async def get_providers(state: StateDep) -> ProvidersResponse:
    """Get available providers and models from agent."""
    providers = await _get_model_providers(state.agent)
    # Build default models map: use first model for each connected provider
    default_models: dict[str, str] = {}
    connected = [p.id for p in providers if any(os.environ.get(env) for env in p.env)]
    for provider in providers:
        if provider.id in connected and provider.models:
            # Simply use the first available model
            default_models[provider.id] = next(iter(provider.models.keys()))

    return ProvidersResponse(providers=providers, default=default_models)


@router.get("/provider")
async def list_providers(state: StateDep) -> ProviderListResponse:
    """List all providers."""
    providers = await _get_model_providers(state.agent)
    # Determine which providers are "connected" based on env vars
    connected = [p.id for p in providers if any(os.environ.get(env) for env in p.env)]
    # Build default models map: use first model for each connected provider
    default_models: dict[str, str] = {}
    for provider in providers:
        if provider.id in connected and provider.models:
            # Simply use the first available model
            default_models[provider.id] = next(iter(provider.models.keys()))

    return ProviderListResponse(all=providers, default=default_models, connected=connected)


@router.get("/mode")
async def list_modes(state: StateDep) -> list[Mode]:
    """List available modes."""
    _ = state  # unused for now
    return [Mode(name="default", tools={})]



"""Codex adapter exceptions."""

from __future__ import annotations

from typing import Any


class CodexError(Exception):
    """Base exception for Codex adapter errors."""


class CodexProcessError(CodexError):
    """Error starting or communicating with the Codex app-server process."""


class CodexRequestError(CodexError):
    """Error from a Codex app-server request (JSON-RPC error response)."""

    def __init__(self, code: int, message: str, data: dict[str, Any] | None = None) -> None:
        super().__init__(message)
        self.code = code
        self.message = message
        self.data = data or {}

    def __str__(self) -> str:
        if self.data:
            return f"[{self.code}] {self.message}: {self.data}"
        return f"[{self.code}] {self.message}"



"""High-level project store with auto-detection."""

from __future__ import annotations

import hashlib
from pathlib import Path
from typing import TYPE_CHECKING

from agentpool.log import get_logger
from agentpool.sessions.models import ProjectData
from agentpool.utils.time_utils import get_now


if TYPE_CHECKING:
    from agentpool.storage.manager import StorageManager

logger = get_logger(__name__)

# Config file names to search for in .agentpool/ directory
CONFIG_FILENAMES = ["config.yml", "config.yaml", "config.json", "config.toml"]


def resolve_config(project: ProjectData | None = None, cwd: str | None = None) -> str | None:
    """Resolve config path using project settings and global fallback.

    Resolution order:
    1. Project's explicit config_path (if set)
    2. .agentpool/config.yml in project worktree (auto-discovered)
    3. Global default from ConfigStore (CLI fallback)
    4. None (use built-in defaults)

    Args:
        project: Project data (optional)
        cwd: Current working directory for discovery if no project

    Returns:
        Path to config file, or None if no config found
    """
    from agentpool_cli.store import ConfigStore

    # 1. Project-specific explicit config
    if project and project.config_path:
        config_path = Path(project.config_path)
        if config_path.is_file():
            return str(config_path)
        logger.warning("Project config not found", path=project.config_path)

    # 2. Auto-discover in project worktree
    if worktree := (project.worktree if project else cwd):
        local_config = discover_config_path(worktree)
        if local_config:
            return local_config

    # 3. Global default from ConfigStore
    try:
        config_store = ConfigStore()
        if active := config_store.get_active():
            config_path = Path(active.path)
            if config_path.is_file():
                return str(config_path)
            logger.warning("Active config not found", path=active.path)
    except Exception:
        logger.exception("Error loading ConfigStore")
    # 4. No config found
    return None


def detect_project_root(cwd: str) -> tuple[str, str | None]:
    """Walk up directory tree to find VCS root or use cwd.

    Args:
        cwd: Current working directory

    Returns:
        Tuple of (worktree_path, vcs_type).
        vcs_type is "git", "hg", or None if no VCS found.
    """
    path = Path(cwd).resolve()
    for parent in [path, *path.parents]:
        if (parent / ".git").exists():
            return str(parent), "git"
        if (parent / ".hg").exists():
            return str(parent), "hg"
    # No VCS found, use the original directory
    return str(path), None


def generate_project_id(worktree: str) -> str:
    """Generate stable hash of canonical worktree path.

    Args:
        worktree: Absolute path to project root

    Returns:
        40-character hex string (SHA1 hash)
    """
    canonical = str(Path(worktree).resolve())
    return hashlib.sha1(canonical.encode()).hexdigest()


def discover_config_path(worktree: str) -> str | None:
    """Search for config file in .agentpool/ directory.

    Args:
        worktree: Project root path

    Returns:
        Path to config file if found, None otherwise
    """
    config_dir = Path(worktree) / ".agentpool"
    if not config_dir.is_dir():
        return None

    for filename in CONFIG_FILENAMES:
        config_path = config_dir / filename
        if config_path.is_file():
            return str(config_path)
    return None


class ProjectStore:
    """High-level API for project management.

    Provides:
    - Auto-detection of project root from cwd
    - Auto-creation of projects
    - Config file discovery
    - Convenient lookup methods
    """

    def __init__(self, storage: StorageManager) -> None:
        """Initialize project store."""
        self.storage = storage

    async def get_or_create(self, cwd: str) -> ProjectData:
        """Get existing project or create one for the given directory.

        Detects VCS root, generates project ID, discovers config,
        and creates/updates the project.

        Args:
            cwd: Current working directory

        Returns:
            ProjectData for the detected/created project
        """
        # Detect project root and VCS
        worktree, vcs = detect_project_root(cwd)

        # Check if project already exists
        if existing := await self.storage.get_project_by_worktree(worktree):
            # Update last_active and return
            await self.storage.touch_project(existing.project_id)
            return existing.touch()

        # Create new project
        project_id = generate_project_id(worktree)
        config_path = discover_config_path(worktree)

        project = ProjectData(
            project_id=project_id,
            worktree=worktree,
            name=None,  # Can be set later
            vcs=vcs,
            config_path=config_path,
            created_at=get_now(),
            last_active=get_now(),
        )

        await self.storage.save_project(project)
        logger.info(
            "Created project",
            project_id=project_id,
            worktree=worktree,
            vcs=vcs,
            config_path=config_path,
        )
        return project

    async def get_by_cwd(self, cwd: str) -> ProjectData | None:
        """Get project for the given directory without creating.

        Args:
            cwd: Current working directory

        Returns:
            ProjectData if found, None otherwise
        """
        worktree, _ = detect_project_root(cwd)
        return await self.storage.get_project_by_worktree(worktree)

    async def get_by_name(self, name: str) -> ProjectData | None:
        """Get project by friendly name.

        Args:
            name: Project name

        Returns:
            ProjectData if found, None otherwise
        """
        return await self.storage.get_project_by_name(name)

    async def get_by_id(self, project_id: str) -> ProjectData | None:
        """Get project by ID.

        Args:
            project_id: Project identifier

        Returns:
            ProjectData if found, None otherwise
        """
        return await self.storage.get_project(project_id)

    async def list_recent(self, limit: int = 20) -> list[ProjectData]:
        """List recent projects ordered by last_active.

        Args:
            limit: Maximum number of projects to return

        Returns:
            List of projects
        """
        projects = await self.storage.list_projects(limit=limit)
        return projects if projects is not None else []

    async def set_name(self, project_id: str, name: str) -> ProjectData | None:
        """Set friendly name for a project.

        Args:
            project_id: Project identifier
            name: New name for the project

        Returns:
            Updated ProjectData, or None if project not found
        """
        project = await self.storage.get_project(project_id)
        if not project:
            return None

        updated = project.model_copy(update={"name": name, "last_active": get_now()})
        await self.storage.save_project(updated)
        return updated

    async def set_config_path(self, project_id: str, config_path: str | None) -> ProjectData | None:
        """Set or clear config path for a project.

        Args:
            project_id: Project identifier
            config_path: Path to config file, or None to use auto-discovery

        Returns:
            Updated ProjectData, or None if project not found
        """
        project = await self.storage.get_project(project_id)
        if not project:
            return None

        updated = project.model_copy(update={"config_path": config_path, "last_active": get_now()})
        await self.storage.save_project(updated)
        return updated

    async def update_settings(self, project_id: str, **settings: object) -> ProjectData | None:
        """Update project-specific settings.

        Args:
            project_id: Project identifier
            **settings: Settings to update

        Returns:
            Updated ProjectData, or None if project not found
        """
        project = await self.storage.get_project(project_id)
        if not project:
            return None

        updated = project.with_settings(**settings)
        await self.storage.save_project(updated)
        return updated

    async def delete(self, project_id: str) -> bool:
        """Delete a project.

        Args:
            project_id: Project identifier

        Returns:
            True if deleted, False if not found
        """
        result = await self.storage.delete_project(project_id)
        return bool(result)

    async def refresh_config(self, project_id: str) -> ProjectData | None:
        """Re-discover config file for a project.

        Useful after user adds .agentpool/config.yml to their project.

        Args:
            project_id: Project identifier

        Returns:
            Updated ProjectData, or None if project not found
        """
        project = await self.storage.get_project(project_id)
        if not project:
            return None

        config_path = discover_config_path(project.worktree)
        if config_path != project.config_path:
            update = {"config_path": config_path, "last_active": get_now()}
            updated = project.model_copy(update=update)
            await self.storage.save_project(updated)
            return updated
        return project



"""Debug toolset for agent self-introspection and runtime debugging."""

from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from datetime import UTC, datetime
import logging
from typing import Any, Literal

import anyenv
from pydantic_ai import RunContext  # noqa: TC002

from agentpool.agents.context import AgentContext  # noqa: TC001
from agentpool.resource_providers import StaticResourceProvider


LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


@dataclass
class LogRecord:
    """A captured log record."""

    level: str
    logger: str
    message: str
    timestamp: str
    extra: dict[str, Any] = field(default_factory=dict)


LEVEL_PRIORITY = {
    "DEBUG": 10,
    "INFO": 20,
    "WARNING": 30,
    "ERROR": 40,
    "CRITICAL": 50,
}


class MemoryLogHandler(logging.Handler):
    """A logging handler that stores records in memory for later retrieval."""

    def __init__(self, max_records: int = 1000) -> None:
        super().__init__()
        self.max_records = max_records
        self.records: deque[LogRecord] = deque(maxlen=max_records)
        self.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))

    def emit(self, record: logging.LogRecord) -> None:
        """Store the log record."""
        log_record = LogRecord(
            level=record.levelname,
            logger=record.name,
            message=self.format(record),
            timestamp=self.formatter.formatTime(record) if self.formatter else "",
            extra={k: v for k, v in record.__dict__.items() if k not in logging.LogRecord.__dict__},
        )
        self.records.append(log_record)

    def get_records(
        self,
        level: LogLevel | None = None,
        logger_filter: str | None = None,
        limit: int | None = None,
    ) -> list[LogRecord]:
        """Get filtered log records.

        Args:
            level: Minimum log level to include
            logger_filter: Only include loggers containing this string
            limit: Maximum number of records to return (newest first)

        Returns:
            List of matching log records
        """
        min_level = LEVEL_PRIORITY.get(level, 0) if level else 0
        filtered = []
        for record in reversed(self.records):
            if LEVEL_PRIORITY.get(record.level, 0) < min_level:
                continue
            if logger_filter and logger_filter not in record.logger:
                continue
            filtered.append(record)
            if limit and len(filtered) >= limit:
                break

        return filtered

    def clear(self) -> None:
        """Clear all stored records."""
        self.records.clear()


# Global memory handler instance
_memory_handler: MemoryLogHandler | None = None


def get_memory_handler() -> MemoryLogHandler:
    """Get or create the global memory log handler."""
    global _memory_handler  # noqa: PLW0603
    if _memory_handler is None:
        _memory_handler = MemoryLogHandler()
        # Attach to root logger to capture everything
        logging.getLogger().addHandler(_memory_handler)
    return _memory_handler


def install_memory_handler(max_records: int = 1000) -> MemoryLogHandler:
    """Install the memory handler if not already installed.

    Args:
        max_records: Maximum number of records to keep in memory

    Returns:
        The memory handler instance
    """
    global _memory_handler  # noqa: PLW0603
    if _memory_handler is None:
        _memory_handler = MemoryLogHandler(max_records=max_records)
        logging.getLogger().addHandler(_memory_handler)
    return _memory_handler


# =============================================================================
# Introspection Tool
# =============================================================================

INTROSPECTION_USAGE = """
Execute Python code with full access to your runtime context.

Available in namespace:
- ctx: AgentContext (ctx.agent, ctx.pool, ctx.config, ctx.definition, etc.)
- run_ctx: pydantic-ai RunContext (current run state)
- me: Shortcut for ctx.agent (your Agent instance)
- save(key, value): Save an object to persist between calls
- state: Dict of saved objects from previous calls

You can inspect yourself, the pool, other agents, your tools, and more.
Write an async main() function that returns the result.

Example - inspect your own tools:
```python
async def main():
    tools = await me.tools.get_tools()
    return [i.name for i in tools]
```

Example - check pool state:
```python
async def main():
    if ctx.pool:
        agents = list(ctx.pool.get_agents().keys())
        return f"Agents in pool: {agents}"
    return "No pool available"
```

Example - explore with dir():
```python
async def main():
    return dir(ctx)
```
"""


# =============================================================================
# Log Tools
# =============================================================================


async def get_logs(
    level: LogLevel = "INFO",
    logger_filter: str | None = None,
    limit: int = 50,
) -> str:
    """Get recent log entries from memory.

    Args:
        level: Minimum log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        logger_filter: Only show logs from loggers containing this string
        limit: Maximum number of log entries to return

    Returns:
        Formatted log entries
    """
    handler = get_memory_handler()
    records = handler.get_records(level=level, logger_filter=logger_filter, limit=limit)
    if not records:
        return "No log entries found matching criteria"

    lines = [f"=== {len(records)} log entries (newest first) ===\n"]
    for record in records:
        lines.append(record.message)  # noqa: PERF401

    return "\n".join(lines)


# =============================================================================
# Path Tools
# =============================================================================


async def get_platform_paths() -> str:
    """Get platform-specific paths for agentpool.

    Returns:
        Dictionary of platform paths
    """
    import platformdirs

    paths = {
        "config": platformdirs.user_config_dir("agentpool"),
        "data": platformdirs.user_data_dir("agentpool"),
        "cache": platformdirs.user_cache_dir("agentpool"),
        "logs": platformdirs.user_log_dir("agentpool"),
        "state": platformdirs.user_state_dir("agentpool"),
    }

    lines = ["Platform paths for agentpool:", ""]
    for name, path in paths.items():
        lines.append(f"  {name}: {path}")

    return "\n".join(lines)


# =============================================================================
# Toolset Class
# =============================================================================


class DebugTools(StaticResourceProvider):
    """Debug and introspection tools for agent development.

    Provides tools for:
    - Self-introspection via code execution with runtime context access
    - Log inspection and management
    - Platform path discovery
    - Stateful namespace for persisting objects between introspection calls
    """

    def __init__(self, name: str = "debug") -> None:
        """Initialize debug tools.

        Args:
            name: Toolset name/namespace
        """
        super().__init__(name=name)
        self._namespace_storage: dict[str, Any] = {}  # Stateful storage for introspection
        desc = (self.execute_introspection.__doc__ or "") + "\n\n" + INTROSPECTION_USAGE
        self._tools = [
            self.create_tool(
                self.execute_introspection, category="other", description_override=desc
            ),
            self.create_tool(get_platform_paths, category="other", read_only=True, idempotent=True),
        ]

    async def execute_introspection(  # noqa: D417
        self, ctx: AgentContext, run_ctx: RunContext[Any], code: str, title: str
    ) -> str:
        """Execute Python code with access to your own runtime context.

        This is a debugging/development tool that gives you full access to
        inspect and interact with your runtime environment.

        Args:
            code: Python code with async main() function to execute
            title: Short descriptive title for this script (3-4 words)

        Returns:
            Result of execution or error message
        """
        # Emit progress with the code being executed
        await ctx.events.tool_call_progress(
            title="Executing introspection code",
            status="in_progress",
            items=[f"```python\n{code}\n```"],
        )

        # Build namespace with runtime context and stateful storage
        def save(key: str, value: Any) -> None:
            """Save a value to persist between introspection calls."""
            self._namespace_storage[key] = value

        namespace: dict[str, Any] = {
            "ctx": ctx,
            "run_ctx": run_ctx,
            "me": ctx.agent,
            "save": save,
            "state": self._namespace_storage.copy(),
        }
        start_time = datetime.now(UTC)
        exit_code = 0
        error_msg = None
        result_str = None
        try:
            exec(code, namespace)
            if "main" not in namespace:
                return "Error: Code must define an async main() function"
            result = await namespace["main"]()
            result_str = (
                str(result)
                if result is not None
                else "Code executed successfully (no return value)"
            )

            await ctx.events.tool_call_progress(
                title="Executed introspection code successfully",
                status="in_progress",
                items=[f"```python\n{code}\n```\n\n```terminal\n{result_str}\n```"],
            )
        except Exception as e:  # noqa: BLE001
            exit_code = 1
            error_msg = f"{type(e).__name__}: {e}"
            result_str = error_msg
        finally:
            # Save to agent's internal filesystem
            timestamp = start_time.strftime("%Y%m%d_%H%M%S")
            # Write script file
            script_path = f"debug/scripts/{timestamp}_{title}.py"
            ctx.internal_fs.pipe(script_path, code.encode())
            # Write metadata file
            metadata = {
                "title": title,
                "timestamp": start_time.isoformat(),
                "exit_code": exit_code,
                "duration": (datetime.now(UTC) - start_time).total_seconds(),
                "result": result_str,
                "error": error_msg,
            }
            metadata_path = f"debug/scripts/{timestamp}_{title}.json"
            ctx.internal_fs.pipe(metadata_path, anyenv.dump_json(metadata, indent=True).encode())
        assert result_str
        return result_str



"""Binary archive extraction utilities."""

from __future__ import annotations

from pathlib import Path
import shutil
import tarfile
import zipfile


def _extract_from_zip(archive_path: Path, binary_name: str, dest_path: Path) -> Path:
    """Extract a binary from a ZIP archive."""
    with zipfile.ZipFile(archive_path, "r") as zip_ref:
        for member in zip_ref.infolist():
            if member.is_dir():
                continue
            if Path(member.filename).name != binary_name:
                continue
            with zip_ref.open(member, "r") as source, dest_path.open("wb") as target:
                shutil.copyfileobj(source, target)
            return dest_path
    raise FileNotFoundError(f"Binary {binary_name!r} not found in ZIP archive {archive_path!r}")


def _extract_from_tar(archive_path: Path, binary_name: str, dest_path: Path) -> Path:
    """Extract a binary from a TAR archive."""
    with tarfile.open(archive_path, "r:*") as tar_ref:
        for member in tar_ref.getmembers():
            if not member.isfile():
                continue
            if Path(member.name).name != binary_name:
                continue
            source = tar_ref.extractfile(member)
            if source is None:
                continue
            with source, dest_path.open("wb") as target:
                shutil.copyfileobj(source, target)
            return dest_path
    raise FileNotFoundError(f"Binary {binary_name!r} not found in TAR archive {archive_path!r}")


def extract_binary(archive_path: Path, binary_name: str, dest_dir: Path) -> Path:
    """Extract a named binary from an archive, or copy it if not an archive.

    Supports ZIP and TAR (gz/bz2/xz) archives. If *archive_path* is not a
    recognised archive format the file is copied directly to *dest_dir*.

    Returns the path to the extracted/copied binary.
    """
    dest_path = dest_dir / binary_name
    dest_dir.mkdir(parents=True, exist_ok=True)

    if zipfile.is_zipfile(archive_path):
        return _extract_from_zip(archive_path, binary_name, dest_path)

    if tarfile.is_tarfile(archive_path):
        return _extract_from_tar(archive_path, binary_name, dest_path)

    # Not an archive — copy directly
    shutil.copy(archive_path, dest_path)
    return dest_path



from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import TYPE_CHECKING, Protocol


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from acp.task import RpcTask


class MessageQueue(Protocol):
    """Protocol for message queues used in RPC task dispatch."""

    async def publish(self, task: RpcTask) -> None: ...

    async def close(self) -> None: ...

    def task_done(self) -> None: ...

    async def join(self) -> None: ...

    def __aiter__(self) -> AsyncIterator[RpcTask]: ...


class InMemoryMessageQueue:
    """Simple in-memory broker for RPC task dispatch."""

    def __init__(self, *, maxsize: int = 0) -> None:
        self._queue: asyncio.Queue[RpcTask | None] = asyncio.Queue(maxsize=maxsize)
        self._closed = False

    async def publish(self, task: RpcTask) -> None:
        if self._closed:
            raise RuntimeError("mssage queue already closed")
        await self._queue.put(task)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._queue.put(None)

    async def join(self) -> None:
        await self._queue.join()

    def task_done(self) -> None:
        with suppress(ValueError):
            self._queue.task_done()

    def __aiter__(self) -> AsyncIterator[RpcTask]:
        return _QueueIterator(self)


class _QueueIterator:
    def __init__(self, queue: InMemoryMessageQueue) -> None:
        self._queue = queue

    def __aiter__(self) -> _QueueIterator:
        return self

    async def __anext__(self) -> RpcTask:
        item = await self._queue._queue.get()
        if item is None:
            self._queue.task_done()
            raise StopAsyncIteration
        return item



"""ClaudeCodeAgent Exceptions."""

from __future__ import annotations


class ThinkingModeAlreadyConfiguredError(ValueError):
    """Raised when attempting to change thinking mode when max_thinking_tokens is configured."""

    def __init__(self) -> None:
        msg = (
            "Cannot change thinking mode: max_thinking_tokens is configured. "
            "The envvar MAX_THINKING_TOKENS takes precedence over the 'ultrathink' keyword."
        )
        super().__init__(msg)



"""Configuration models for observability providers."""

from __future__ import annotations

from typing import Annotated, Any, Literal

from pydantic import ConfigDict, Field, PrivateAttr, SecretStr
from schemez import Schema


class BaseObservabilityConfig(Schema):
    """Base configuration for observability endpoints."""

    type: str = Field(title="Observability provider type")
    """Observability provider type."""

    enabled: bool = Field(default=True, title="Provider enabled")
    """Provider enabled state."""

    service_name: str | None = Field(
        default=None,
        examples=["agentpool", "my-ai-service", "production-bot"],
        title="Service name",
    )
    environment: str | None = Field(
        default=None,
        examples=["production", "staging", "development"],
        title="Environment",
    )
    protocol: Literal["http/protobuf", "grpc", "http/json"] = Field(
        default="http/protobuf",
        examples=["http/protobuf", "grpc", "http/json"],
        title="Protocol",
    )


class LogfireObservabilityConfig(BaseObservabilityConfig):
    """Configuration for Logfire endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "Logfire"})

    type: Literal["logfire"] = "logfire"
    """Logfire endpoint type."""

    token: SecretStr | None = Field(default=None, title="Logfire token")
    """Logfire token."""

    region: Literal["us", "eu"] = Field(default="us", examples=["us", "eu"], title="Region")
    """Logfire region."""

    # Private - computed from region
    _endpoint: str = PrivateAttr()
    _headers: dict[str, str] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any, /) -> None:
        """Compute private attributes from user config."""
        endpoint = (
            "https://logfire-eu.pydantic.dev"
            if self.region == "eu"
            else "https://logfire-us.pydantic.dev"
        )
        object.__setattr__(self, "_endpoint", endpoint)

        if self.token:
            headers = {"Authorization": f"Bearer {self.token.get_secret_value()}"}
            object.__setattr__(self, "_headers", headers)


class LangsmithObservabilityConfig(BaseObservabilityConfig):
    """Configuration for Langsmith endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "Langsmith"})

    type: Literal["langsmith"] = "langsmith"
    """Langsmith observability configuration."""

    api_key: SecretStr | None = Field(default=None, title="Langsmith API key")
    """Langsmith API key."""

    project_name: str | None = Field(
        default=None,
        examples=["my-project", "ai-agents", "production"],
        title="Project name",
    )

    _endpoint: str = PrivateAttr(default="https://api.smith.langchain.com")
    _headers: dict[str, str] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any, /) -> None:
        """Compute private attributes from user config."""
        if self.api_key:
            headers = {"x-api-key": self.api_key.get_secret_value()}
            object.__setattr__(self, "_headers", headers)


class AgentOpsObservabilityConfig(BaseObservabilityConfig):
    """Configuration for AgentOps endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "AgentOps"})

    type: Literal["agentops"] = "agentops"
    """AgentOps observability configuration."""

    api_key: SecretStr | None = Field(default=None, title="AgentOps API key")
    """AgentOps API key."""

    _endpoint: str = PrivateAttr(default="https://api.agentops.ai")
    _headers: dict[str, str] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any, /) -> None:
        """Compute private attributes from user config."""
        if self.api_key:
            headers = {"Authorization": f"Bearer {self.api_key.get_secret_value()}"}
            object.__setattr__(self, "_headers", headers)


class ArizePhoenixObservabilityConfig(BaseObservabilityConfig):
    """Configuration for Arize Phoenix endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "Arize Phoenix"})

    type: Literal["arize"] = "arize"
    """Arize observability configuration."""

    api_key: SecretStr | None = Field(default=None, title="Arize API key")
    """Arize API key."""

    space_key: str | None = Field(
        default=None,
        examples=["default", "team-space", "production-space"],
        title="Space key",
    )
    """Arize space key."""

    model_id: str | None = Field(
        default=None,
        examples=["gpt-4", "claude-3", "my-model"],
        title="Model ID",
    )
    """Arize model ID."""

    _endpoint: str = PrivateAttr(default="https://api.arize.com")
    _headers: dict[str, str] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any, /) -> None:
        """Compute private attributes from user config."""
        if self.api_key:
            headers = {"Authorization": f"Bearer {self.api_key.get_secret_value()}"}
            object.__setattr__(self, "_headers", headers)


class AxiomObservabilityConfig(BaseObservabilityConfig):
    """Configuration for Axiom endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "Axiom"})

    type: Literal["axiom"] = "axiom"
    """Axiom observability configuration."""

    api_token: SecretStr | None = Field(default=None, title="Axiom API token")
    """Axiom API token with ingest permissions."""

    dataset: str = Field(
        examples=["traces", "otel-traces", "my-service-traces"],
        title="Dataset name",
    )
    """Axiom dataset name where traces are sent."""

    region: Literal["us", "eu"] | None = Field(
        default=None,
        examples=["us", "eu"],
        title="Region",
    )
    """Axiom region. If not set, uses default cloud endpoint."""

    _endpoint: str = PrivateAttr()
    _headers: dict[str, str] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any, /) -> None:
        """Compute private attributes from user config."""
        # Axiom uses /v1/traces endpoint for OTLP
        if self.region == "eu":
            endpoint = "https://api.eu.axiom.co/v1/traces"
        else:
            endpoint = "https://api.axiom.co/v1/traces"
        object.__setattr__(self, "_endpoint", endpoint)

        headers: dict[str, str] = {}
        if self.api_token:
            headers["Authorization"] = f"Bearer {self.api_token.get_secret_value()}"
        headers["X-Axiom-Dataset"] = self.dataset
        object.__setattr__(self, "_headers", headers)


class CustomObservabilityConfig(BaseObservabilityConfig):
    """Configuration for custom OTEL endpoint."""

    model_config = ConfigDict(json_schema_extra={"x-doc-title": "Custom OTEL"})

    type: Literal["custom"] = "custom"
    """Custom OTEL endpoint configuration."""

    endpoint: str = Field(
        examples=["https://otel.example.com", "http://localhost:4318"],
        title="OTEL endpoint",
    )
    """Custom OTEL endpoint URL."""

    headers: dict[str, str] = Field(default_factory=dict, title="Custom headers")
    """Custom headers for the OTEL endpoint."""


# Union of all provider configs
ObservabilityProviderConfig = Annotated[
    LogfireObservabilityConfig
    | LangsmithObservabilityConfig
    | AgentOpsObservabilityConfig
    | ArizePhoenixObservabilityConfig
    | AxiomObservabilityConfig
    | CustomObservabilityConfig,
    Field(discriminator="type"),
]


class ObservabilityConfig(Schema):
    """Global observability configuration - supports single backend only."""

    enabled: bool = Field(default=True, title="Observability enabled")
    """Whether observability is enabled."""

    provider: ObservabilityProviderConfig | None = Field(
        default=None,
        title="Observability provider",
    )
    """Single observability provider configuration."""



"""Base resource provider interface."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Self

from anyenv.signals import Signal

from agentpool.log import get_logger
from agentpool.tools.base import Tool
from agentpool_config.tools import ToolHints


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from types import TracebackType

    from pydantic_ai import ModelRequestPart
    from schemez import OpenAIFunctionDefinition

    from agentpool.prompts.prompts import BasePrompt
    from agentpool.resource_providers.resource_info import ResourceInfo
    from agentpool.skills.skill import Skill
    from agentpool.tools.base import ToolKind


logger = get_logger(__name__)


ResourceType = Literal["tools", "prompts", "resources", "skills"]
ProviderKind = Literal[
    "base", "mcp", "mcp_run", "tools", "prompts", "skills", "aggregating", "custom"
]


@dataclass(frozen=True, slots=True)
class ResourceChangeEvent:
    """Event emitted when resources change in a provider.

    Attributes:
        provider_name: Name of the provider instance
        provider_kind: Kind/type of the provider (e.g., "mcp", "tools")
        resource_type: Type of resource that changed
        owner: Optional owner of the provider (e.g., agent name)
    """

    provider_name: str
    provider_kind: ProviderKind
    resource_type: ResourceType
    owner: str | None = None


class ResourceProvider:
    """Base class for resource providers.

    Provides tools, prompts, and other resources to agents.
    Default implementations return empty lists - override as needed.

    Class Attributes:
        kind: Short slug identifying the provider type (e.g., "mcp", "tools")

    Change signals (using anyenv.signals.Signal):
        - tools_changed: Emitted when tools change
        - prompts_changed: Emitted when prompts change
        - resources_changed: Emitted when resources change
        - skills_changed: Emitted when skills change

    Example:
        provider.tools_changed.connect(my_handler)
        await provider.tools_changed.emit(provider.create_change_event("tools"))
    """

    kind: ProviderKind = "base"

    # Change signals - emit ResourceChangeEvent when resources change
    tools_changed: Signal[ResourceChangeEvent] = Signal()
    prompts_changed: Signal[ResourceChangeEvent] = Signal()
    resources_changed: Signal[ResourceChangeEvent] = Signal()
    skills_changed: Signal[ResourceChangeEvent] = Signal()

    def __init__(self, name: str, owner: str | None = None) -> None:
        """Initialize the resource provider."""
        self.name = name
        self.owner = owner
        self.log = logger.bind(name=self.name, owner=self.owner)

    def create_change_event(self, resource_type: ResourceType) -> ResourceChangeEvent:
        """Create a ResourceChangeEvent for this provider."""
        return ResourceChangeEvent(
            provider_name=self.name,
            provider_kind=self.kind,
            resource_type=resource_type,
            owner=self.owner,
        )

    async def __aenter__(self) -> Self:
        """Async context entry if required."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async context cleanup if required."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r})"

    async def get_tools(self) -> Sequence[Tool]:
        """Get available tools. Override to provide tools."""
        return []

    async def get_tool(self, tool_name: str) -> Tool:
        """Get specific tool."""
        tools = await self.get_tools()
        for tool in tools:
            if tool.name == tool_name:
                return tool

        raise ValueError(f"Tool {tool_name!r} not found")

    async def get_prompts(self) -> list[BasePrompt]:
        """Get available prompts. Override to provide prompts."""
        return []

    async def get_resources(self) -> list[ResourceInfo]:
        """Get available resources. Override to provide resources."""
        return []

    async def get_skills(self) -> list[Skill]:
        """Get available skills. Override to provide skills."""
        return []

    async def get_skill_instructions(self, skill_name: str) -> str:
        """Get full instructions for a specific skill.

        Args:
            skill_name: Name of the skill to get instructions for

        Returns:
            The full skill instructions for execution

        Raises:
            KeyError: If skill not found
        """
        raise KeyError(f"Skill {skill_name!r} not found")

    async def get_request_parts(
        self, name: str, arguments: dict[str, str] | None = None
    ) -> list[ModelRequestPart]:
        """Get a prompt formatted with arguments.

        Args:
            name: Name of the prompt to format
            arguments: Optional arguments for prompt formatting

        Returns:
            Single chat message with merged content

        Raises:
            KeyError: If prompt not found
            ValueError: If formatting fails
        """
        prompts = await self.get_prompts()
        prompt = next((p for p in prompts if p.name == name), None)
        if not prompt:
            raise KeyError(f"Prompt {name!r} not found")

        messages = await prompt.format(arguments or {})
        if not messages:
            raise ValueError(f"Prompt {name!r} produced no messages")

        return [p for prompt_msg in messages for p in prompt_msg.to_pydantic_parts()]

    def create_tool(
        self,
        fn: Callable[..., Any],
        read_only: bool | None = None,
        destructive: bool | None = None,
        idempotent: bool | None = None,
        open_world: bool | None = None,
        requires_confirmation: bool = False,
        metadata: dict[str, Any] | None = None,
        category: ToolKind | None = None,
        name_override: str | None = None,
        description_override: str | None = None,
        schema_override: OpenAIFunctionDefinition | None = None,
    ) -> Tool:
        """Create a tool from a function.

        Args:
            fn: Function to create a tool from
            read_only: Whether the tool is read-only
            destructive: Whether the tool is destructive
            idempotent: Whether the tool is idempotent
            open_world: Whether the tool is open-world
            requires_confirmation: Whether the tool requires confirmation
            metadata: Metadata for the tool
            category: Category of the tool
            name_override: Override the name of the tool
            description_override: Override the description of the tool
            schema_override: Override the schema of the tool

        Returns:
            Tool created from the function
        """
        return Tool.from_callable(
            fn=fn,
            category=category,
            source=self.name,
            requires_confirmation=requires_confirmation,
            metadata=metadata,
            name_override=name_override,
            description_override=description_override,
            schema_override=schema_override,
            hints=ToolHints(
                read_only=read_only,
                destructive=destructive,
                idempotent=idempotent,
                open_world=open_world,
            ),
        )



"""Message container with statistics and formatting capabilities."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal

from psygnal.containers import EventedList

from agentpool.log import get_logger
from agentpool.messaging import ChatMessage
from agentpool.messaging.chat_filesystem import ChatMessageFileSystem
from agentpool.utils.count_tokens import batch_count_tokens


if TYPE_CHECKING:
    from datetime import datetime

    from agentpool.common_types import MessageRole
    from agentpool.messaging.messages import FormatStyle
    from agentpool.utils.dag import DAGNode


logger = get_logger(__name__)
DEFAULT_TOKEN_MODEL = "gpt-4.1"


class ChatMessageList(EventedList[ChatMessage[Any]]):
    """Container for tracking and managing chat messages.

    Extends EventedList to provide:
    - Message statistics (tokens, costs)
    - History formatting
    - Token-aware context window management
    - Role-based filtering
    """

    def get_history_tokens(self, fallback_model: str | None = None) -> int:
        """Get total token count for all messages.

        Uses cost_info when available, falls back to tiktoken estimation
        for messages without usage information.

        Returns:
            Total token count across all messages
        """
        # Use cost_info if available
        total = sum(m.cost_info.token_usage.total_tokens for m in self if m.cost_info)
        # For messages without cost_info, estimate using tiktoken
        if msgs := [msg for msg in self if not msg.cost_info]:
            if fallback_model:
                model_name = fallback_model
            else:
                model_name = next((m.model_name for m in self if m.model_name), DEFAULT_TOKEN_MODEL)
            contents = [str(msg.content) for msg in msgs]
            total += sum(batch_count_tokens(contents, model_name))

        return total

    @cached_property
    def fs(self) -> ChatMessageFileSystem:
        return ChatMessageFileSystem(self)

    def get_total_cost(self) -> float:
        """Calculate total cost in USD across all messages.

        Only includes messages with cost information.

        Returns:
            Total cost in USD
        """
        return sum(float(msg.cost_info.total_cost) for msg in self if msg.cost_info)

    @property
    def last_message(self) -> ChatMessage[Any] | None:
        """Get most recent message or None if empty."""
        return self[-1] if self else None

    def format(self, *, style: FormatStyle = "simple", **kwargs: Any) -> str:
        """Format conversation history with configurable style.

        Args:
            style: Formatting style to use
            **kwargs: Additional formatting options passed to message.format()

        Returns:
            Formatted conversation history as string
        """
        return "\n".join(msg.format(style=style, **kwargs) for msg in self)

    def filter_by_role(
        self,
        role: MessageRole,
        *,
        max_messages: int | None = None,
    ) -> list[ChatMessage[Any]]:
        """Get messages with specific role.

        Args:
            role: Role to filter by (user/assistant/system)
            max_messages: Optional limit on number of messages to return

        Returns:
            List of messages with matching role
        """
        messages = [msg for msg in self if msg.role == role]
        return messages[-max_messages:] if max_messages else messages

    def get_between(
        self,
        *,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
    ) -> list[ChatMessage[Any]]:
        """Get messages within a time range.

        Args:
            start_time: Optional start of range
            end_time: Optional end of range

        Returns:
            List of messages within the time range
        """
        messages = list(self)
        if start_time:
            messages = [msg for msg in messages if msg.timestamp >= start_time]
        if end_time:
            messages = [msg for msg in messages if msg.timestamp <= end_time]
        return messages

    def _build_flow_dag(self, message: ChatMessage[Any]) -> DAGNode | None:
        """Build DAG from message flow.

        Args:
            message: Message to build flow DAG for

        Returns:
            Root DAGNode of the graph, or None

        Note:
            forwarded_from has been removed. This method now returns None.
            Flow tracking can be reconstructed from parent_id chain or pool history.
        """
        return None

    def to_mermaid_graph(
        self,
        message: ChatMessage[Any],
        *,
        title: str = "",
        theme: str | None = None,
        rankdir: Literal["TB", "BT", "LR", "RL"] = "LR",
    ) -> str:
        """Convert message flow to mermaid graph."""
        from agentpool.utils.dag import dag_to_list

        dag = self._build_flow_dag(message)
        if not dag:
            return ""

        connections = dag_to_list(dag)

        # Convert to mermaid
        lines = ["```mermaid"]
        if title:
            lines.extend(["---", f"title: {title}", "---"])
        if theme:
            lines.append(f'%%{{ init: {{ "theme": "{theme}" }} }}%%')
        lines.append(f"flowchart {rankdir}")

        # Add connections
        for parent, child in connections:
            lines.append(f"    {parent}-->{child}")

        lines.append("```")
        return "\n".join(lines)



"""Task package."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal


RpcTaskKind = Literal["request", "notification"]


@dataclass(slots=True)
class RpcTask:
    """RpcTask represents a task to be executed by the agent."""

    kind: RpcTaskKind
    message: dict[str, Any]


from .dispatcher import (
    DefaultMessageDispatcher,
    MessageDispatcher,
    NotificationRunner,
    RequestRunner,
)
from .queue import InMemoryMessageQueue, MessageQueue
from .sender import MessageSender, SenderFactory
from .state import InMemoryMessageStateStore, MessageStateStore
from .supervisor import TaskSupervisor
from .debug import DebugEntry, DebuggingMessageStateStore

__all__ = [
    "DebugEntry",
    "DebuggingMessageStateStore",
    "DefaultMessageDispatcher",
    "InMemoryMessageQueue",
    "InMemoryMessageStateStore",
    "MessageDispatcher",
    "MessageQueue",
    "MessageSender",
    "MessageStateStore",
    "NotificationRunner",
    "RequestRunner",
    "RpcTask",
    "RpcTaskKind",
    "SenderFactory",
    "TaskSupervisor",
]



"""Codex data types."""

from __future__ import annotations

from typing import Annotated, Any, Literal

from pydantic import BaseModel, Discriminator, Field, Tag

from codex_adapter.models.base import CodexBaseModel


# Type aliases for Codex types
ModelProvider = Literal["openai", "anthropic", "google", "mistral"]
ReasoningEffort = Literal["none", "minimal", "low", "medium", "high", "xhigh"]
ReasoningSummary = Literal["auto", "concise", "detailed", "none"]
ApprovalPolicy = Literal["untrusted", "on-failure", "on-request", "never"]
SandboxMode = Literal["read-only", "workspace-write", "danger-full-access"]
NetworkAccess = Literal["restricted", "enabled"]
Personality = Literal["none", "friendly", "pragmatic"]
TurnStatus = Literal["pending", "inProgress", "completed", "error", "interrupted"]
ItemType = Literal[
    "reasoning",
    "agent_message",
    "command_execution",
    "user_message",
    "file_change",
    "mcp_tool_call",
]
ItemStatus = Literal["pending", "running", "completed", "error"]

# New type aliases
SessionSource = Literal["cli", "vscode", "exec", "appServer", "unknown"]
ThreadSortKey = Literal["created_at", "updated_at"]
ThreadSourceKind = Literal[
    "cli",
    "vscode",
    "exec",
    "appServer",
    "subAgent",
    "subAgentReview",
    "subAgentCompact",
    "subAgentThreadSpawn",
    "subAgentOther",
    "unknown",
]
MessagePhase = Literal["commentary", "final_answer"]
PatchApplyStatus = Literal["inProgress", "completed", "failed", "declined"]
CommandExecutionStatus = Literal["inProgress", "completed", "failed", "declined"]
McpToolCallStatus = Literal["inProgress", "completed", "failed"]
DynamicToolCallStatus = Literal["inProgress", "completed", "failed"]
CollabAgentTool = Literal["spawnAgent", "sendInput", "resumeAgent", "wait", "closeAgent"]
CollabAgentToolCallStatus = Literal["inProgress", "completed", "failed"]
CollabAgentStatus = Literal[
    "pendingInit", "running", "completed", "errored", "shutdown", "notFound"
]
InputModality = Literal["text", "image"]
SkillScope = Literal["user", "repo", "system", "admin"]
McpAuthStatusValue = Literal["Unsupported", "NotAuthenticated", "Authenticated"]
ReviewDelivery = Literal["inline", "detached"]
ThreadActiveFlag = Literal["waitingOnApproval", "waitingOnUserInput"]
CommandExecutionApprovalDecision = Literal["allow", "allowForSession", "deny", "denyForSession"]
FileChangeApprovalDecision = Literal["allow", "allowForSession", "deny", "denyForSession"]
SkillApprovalDecision = Literal["allow", "deny"]
ModelRerouteReason = Literal["rateLimited", "contextWindowExceeded", "other"]
WriteStatus = Literal["ok", "conflict"]
MergeStrategy = Literal["replace", "merge"]
ExperimentalFeatureStage = Literal["alpha", "beta"]
ElicitationAction = Literal["accept", "decline", "cancel"]
NetworkApprovalProtocol = Literal["http", "https", "socks5Tcp", "socks5Udp"]
NetworkPolicyRuleAction = Literal["allow", "deny"]
ExternalAgentConfigMigrationItemType = Literal["AGENTS_MD", "CONFIG", "SKILLS", "MCP_SERVER_CONFIG"]
PlanType = Literal["free", "go", "plus", "pro", "team", "business", "enterprise", "edu", "unknown"]


# ============================================================================
# AskForApproval (tagged union: string literals or {"reject": RejectConfig})
# ============================================================================


class RejectConfig(CodexBaseModel):
    """Fine-grained rejection controls for approval prompts.

    When a field is True, prompts of that category are automatically
    rejected instead of shown to the user.
    """

    sandbox_approval: bool
    rules: bool
    mcp_elicitations: bool


class RejectApprovalPolicy(CodexBaseModel):
    """Approval policy variant with fine-grained rejection controls."""

    reject: RejectConfig


def _ask_for_approval_discriminator(v: Any) -> str:
    match v:
        case str():
            return "simple"
        case {"reject": _}:
            return "reject"
        case RejectApprovalPolicy():
            return "reject"
        case _:
            return "simple"


AskForApproval = Annotated[
    Annotated[ApprovalPolicy, Tag("simple")] | Annotated[RejectApprovalPolicy, Tag("reject")],
    Discriminator(_ask_for_approval_discriminator),
]
"""Full AskForApproval type: simple string policy or reject config."""


# ============================================================================
# SandboxPolicy (discriminated union on 'type' field)
# ============================================================================


# Mapping from camelCase type values (turn-level API) to kebab-case (thread-level API)
_SANDBOX_TYPE_ALIASES: dict[str, str] = {
    "workspaceWrite": "workspace-write",
    "dangerFullAccess": "danger-full-access",
    "readOnly": "read-only",
    "externalSandbox": "external-sandbox",
}

_READ_ONLY_ACCESS_TYPE_ALIASES: dict[str, str] = {
    "fullAccess": "full-access",
}


def _sandbox_policy_discriminator(v: Any) -> str:
    match v:
        case {"type": str(raw_type)}:
            return _SANDBOX_TYPE_ALIASES.get(raw_type, raw_type)
        case BaseModel():
            return str(v.model_fields["type"].default)
        case _:
            return str(v)


def _read_only_access_discriminator(v: Any) -> str:
    match v:
        case {"type": str(raw_type)}:
            return _READ_ONLY_ACCESS_TYPE_ALIASES.get(raw_type, raw_type)
        case BaseModel():
            return str(v.model_fields["type"].default)
        case _:
            return str(v)


class RestrictedReadOnlyAccess(CodexBaseModel):
    """Restrict reads to an explicit set of roots."""

    type: Literal["restricted"]
    readable_roots: list[str] = Field(default_factory=list)
    include_platform_defaults: bool = True


class FullAccessReadOnlyAccess(CodexBaseModel):
    """Allow unrestricted file reads."""

    type: Literal["full-access", "fullAccess"]


ReadOnlyAccess = Annotated[
    Annotated[RestrictedReadOnlyAccess, Tag("restricted")]
    | Annotated[FullAccessReadOnlyAccess, Tag("full-access")],
    Discriminator(_read_only_access_discriminator),
]


class DangerFullAccessSandboxPolicy(CodexBaseModel):
    """No restrictions whatsoever. Use with caution."""

    type: Literal["danger-full-access", "dangerFullAccess"]


class ReadOnlySandboxPolicy(CodexBaseModel):
    """Read-only access configuration."""

    type: Literal["read-only", "readOnly"]
    access: ReadOnlyAccess | None = None


class ExternalSandboxPolicy(CodexBaseModel):
    """Process is already in an external sandbox."""

    type: Literal["external-sandbox", "externalSandbox"]
    network_access: NetworkAccess = "restricted"


class WorkspaceWriteSandboxPolicy(CodexBaseModel):
    """Grants write access to the workspace directory."""

    type: Literal["workspace-write", "workspaceWrite"]
    writable_roots: list[str] = Field(default_factory=list)
    read_only_access: ReadOnlyAccess | None = None
    network_access: bool = False
    exclude_slash_tmp: bool = False
    exclude_tmpdir_env_var: bool = False


SandboxPolicy = Annotated[
    Annotated[DangerFullAccessSandboxPolicy, Tag("danger-full-access")]
    | Annotated[ReadOnlySandboxPolicy, Tag("read-only")]
    | Annotated[ExternalSandboxPolicy, Tag("external-sandbox")]
    | Annotated[WorkspaceWriteSandboxPolicy, Tag("workspace-write")],
    Discriminator(_sandbox_policy_discriminator),
]
"""Discriminated union for sandbox execution restrictions."""



"""Utilities for importing callables and classes from dotted paths."""

from __future__ import annotations

import importlib
import inspect
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from collections.abc import Callable


def import_callable(path: str) -> Callable[..., Any]:
    """Import a callable from a dotted path.

    Supports both dot and colon notation:
    - Dot notation: module.submodule.Class.method
    - Colon notation: module.submodule:Class.method

    Examples:
        >>> import_callable("os.path.join")
        >>> import_callable("builtins.str.upper")
        >>> import_callable("sqlalchemy.orm:Session.query")

    Args:
        path: Import path using dots and/or colon

    Returns:
        Imported callable

    Raises:
        ValueError: If path cannot be imported or result isn't callable
    """
    if not path:
        raise ValueError("Import path cannot be empty")

    # Normalize path - replace colon with dot if present
    normalized_path = path.replace(":", ".")
    parts = normalized_path.split(".")

    # Try importing progressively smaller module paths
    for i in range(len(parts), 0, -1):
        try:
            # Try current module path
            module_path = ".".join(parts[:i])
            module = importlib.import_module(module_path)

            # Walk remaining parts as attributes
            obj = module
            for part in parts[i:]:
                obj = getattr(obj, part)

            # Check if we got a callable
            if callable(obj):
                return obj
            raise ValueError(f"Found object at {path} but it isn't callable")
        except ImportError:
            # Try next shorter path
            continue
        except AttributeError:
            # Attribute not found - try next shorter path
            continue
    # If we get here, no import combination worked
    raise ValueError(f"Could not import callable from path: {path}")


def import_class(path: str) -> type:
    """Import a class from a dotted path.

    Args:
        path: Dot-separated path to the class

    Returns:
        The imported class

    Raises:
        ValueError: If path is invalid or doesn't point to a class
    """
    try:
        obj = import_callable(path)
        if not isinstance(obj, type):
            raise TypeError(f"{path} is not a class")  # noqa: TRY301
    except Exception as exc:
        raise ValueError(f"Failed to import class from {path}") from exc
    else:
        return obj


if __name__ == "__main__":
    # ATTENTION: Dont modify this script.
    import sys

    if len(sys.argv) != 2:  # noqa: PLR2004
        print("Usage: python importing.py <dot.path.to.object>", file=sys.stderr)
        sys.exit(1)

    dot_path = sys.argv[1]

    try:
        obj = import_callable(dot_path)
        source = inspect.getsource(obj)
        print(source)
    except Exception as e:  # noqa: BLE001
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)



"""Base class for text sharing services."""

from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal


if TYPE_CHECKING:
    from agentpool.messaging.message_history import MessageHistory


Visibility = Literal["public", "unlisted", "private"]


@dataclass
class ShareResult:
    """Result of a text sharing operation."""

    url: str
    """URL to access the shared content."""

    raw_url: str | None = None
    """Direct URL to raw content, if available."""

    delete_url: str | None = None
    """URL to delete the content, if available."""

    id: str | None = None
    """Provider-specific ID of the shared content."""


class TextSharer(abc.ABC):
    """Base class for text sharing services."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of the sharing service."""

    @abc.abstractmethod
    async def share(
        self,
        content: str,
        *,
        title: str | None = None,
        syntax: str | None = None,
        visibility: Visibility = "unlisted",
        expires_in: int | None = None,
    ) -> ShareResult:
        """Share text content.

        Args:
            content: The text content to share
            title: Optional title/filename for the content
            syntax: Syntax highlighting hint (e.g. "python", "markdown")
            visibility: Visibility level (not all providers support all levels)
            expires_in: Expiration time in seconds (not supported by all providers)

        Returns:
            ShareResult with URL and metadata
        """

    async def share_conversation(
        self,
        conversation: MessageHistory,
        *,
        title: str | None = None,
        visibility: Visibility = "unlisted",
        expires_in: int | None = None,
        num_messages: int | None = None,
    ) -> ShareResult:
        """Share a conversation in structured format.

        Default implementation formats conversation as text and calls share().
        Providers that support structured conversations (e.g., OpenCode)
        should override this method.

        Args:
            conversation: MessageHistory object to share
            title: Optional title for the conversation
            visibility: Visibility level
            expires_in: Expiration time in seconds
            num_messages: Number of messages to include (None = all)

        Returns:
            ShareResult with URL and metadata
        """
        # Default: format as plain text
        content = await conversation.format_history(
            num_messages=num_messages,
        )
        return await self.share(
            content,
            title=title,
            syntax="markdown",
            visibility=visibility,
            expires_in=expires_in,
        )

    def share_sync(
        self,
        content: str,
        *,
        title: str | None = None,
        syntax: str | None = None,
        visibility: Visibility = "unlisted",
        expires_in: int | None = None,
    ) -> ShareResult:
        """Synchronous version of share."""
        from anyenv import run_sync

        async def wrapper() -> ShareResult:
            return await self.share(
                content,
                title=title,
                syntax=syntax,
                visibility=visibility,
                expires_in=expires_in,
            )

        return run_sync(wrapper())



"""Helper functions for ACP agent."""

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Sequence

    import structlog

    from acp.schema import AgentCapabilities
    from acp.schema.mcp import McpServer


def filter_servers_by_capabilities(
    servers: Sequence[McpServer],
    agent_capabilities: AgentCapabilities | None,
    *,
    logger: structlog.stdlib.BoundLogger | None = None,
) -> list[McpServer]:
    """Filter MCP servers based on agent's supported transports.

    Args:
        servers: ACP-schema MCP servers to filter (not agentpool_config types)
        agent_capabilities: Agent's capabilities (None = no HTTP/SSE support)
        logger: Optional logger for warnings about unsupported servers

    Returns:
        Servers compatible with the agent's capabilities
    """
    from acp.schema.mcp import HttpMcpServer, SseMcpServer

    # Check what transports are supported
    supports_http = (
        agent_capabilities
        and agent_capabilities.mcp_capabilities
        and agent_capabilities.mcp_capabilities.http
    )
    supports_sse = (
        agent_capabilities
        and agent_capabilities.mcp_capabilities
        and agent_capabilities.mcp_capabilities.sse
    )

    supported_servers: list[McpServer] = []
    unsupported_servers: list[tuple[McpServer, str]] = []

    for server in servers:
        match server:
            case HttpMcpServer() if not supports_http:
                unsupported_servers.append((server, "HTTP"))
            case SseMcpServer() if not supports_sse:
                unsupported_servers.append((server, "SSE"))
            case _:
                # Stdio servers or supported transport types
                supported_servers.append(server)

    # Log warning if some servers were filtered out
    if unsupported_servers and logger:
        transports = ", ".join(sorted({t for _, t in unsupported_servers}))
        server_names = ", ".join(s.name for s, _ in unsupported_servers)
        logger.warning(
            "Agent does not support some MCP transports, skipping servers",
            unsupported_transports=transports,
            skipped_servers=server_names,
            unsupported_count=len(unsupported_servers),
            supported_http=supports_http,
            supported_sse=supports_sse,
        )

    return supported_servers



"""List directory tool implementation."""

from __future__ import annotations

from dataclasses import dataclass
from fnmatch import fnmatch
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any

from upathtools import is_directory

from agentpool.agents.context import AgentContext  # noqa: TC001
from agentpool.log import get_logger
from agentpool.tools.base import Tool, ToolResult


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from exxec import ExecutionEnvironment
    from fsspec.asyn import AsyncFileSystem


logger = get_logger(__name__)


@dataclass
class ListDirectoryTool(Tool[ToolResult]):
    """List files in a directory with filtering support.

    A standalone tool for listing directory contents with:
    - Glob pattern matching (*, **, *.py, etc.)
    - Exclude patterns for filtering
    - Recursive directory traversal with depth control
    - Formatted markdown output

    Use create_list_directory_tool() factory for convenient instantiation.
    """

    # Tool-specific configuration
    env: ExecutionEnvironment | None = None
    """Execution environment to use. Falls back to agent.env if not set."""

    cwd: str | None = None
    """Working directory for resolving relative paths."""

    max_items: int = 500
    """Maximum number of items to return (safety limit)."""

    def get_callable(self) -> Callable[..., Awaitable[ToolResult]]:
        """Return the list_directory method as the callable."""
        return self._list_directory

    def _get_fs(self, ctx: AgentContext) -> AsyncFileSystem:
        """Get filesystem from env, falling back to agent's env if not set."""
        return ctx.agent.env.get_fs() if self.env is None else self.env.get_fs()

    def _resolve_path(self, path: str, ctx: AgentContext) -> str:
        """Resolve a potentially relative path to an absolute path."""
        cwd: str | None = None
        if self.cwd:
            cwd = self.cwd
        elif self.env and self.env.cwd:
            cwd = self.env.cwd
        elif ctx.agent.env and ctx.agent.env.cwd:
            cwd = ctx.agent.env.cwd

        if cwd and not (path.startswith("/") or (len(path) > 1 and path[1] == ":")):
            return str(Path(cwd) / path)
        return path

    async def _list_directory(
        self,
        ctx: AgentContext,
        path: str,
        *,
        pattern: str = "*",
        exclude: list[str] | None = None,
        max_depth: int = 1,
    ) -> ToolResult:
        """List files in a directory with filtering support.

        Args:
            ctx: Agent context for event emission and filesystem access
            path: Base directory to list
            pattern: Glob pattern to match files against. Use "*.py" to match Python
                files in current directory only, or "**/*.py" to match recursively.
                The max_depth parameter limits how deep "**" patterns search.
            exclude: List of patterns to exclude (uses fnmatch against relative paths)
            max_depth: Maximum directory depth to search (default: 1 = current dir only).
                Only affects recursive "**" patterns.

        Returns:
            Markdown-formatted directory listing
        """
        from agentpool_toolsets.fsspec_toolset.helpers import format_directory_listing

        path = self._resolve_path(path, ctx)
        msg = f"Listing directory: {path}"
        await ctx.events.tool_call_start(title=msg, kind="read", locations=[path])

        try:
            fs = self._get_fs(ctx)
            # Check if path exists
            if not await fs._exists(path):
                error_msg = f"Path does not exist: {path}"
                await ctx.events.file_operation("list", path=path, success=False, error=error_msg)
                return ToolResult(
                    content=f"Error: {error_msg}",
                    metadata={"count": 0, "truncated": False},
                )

            # Build glob path
            glob_pattern = f"{path.rstrip('/')}/{pattern}"
            paths = await fs._glob(glob_pattern, maxdepth=max_depth, detail=True)

            files: list[dict[str, Any]] = []
            dirs: list[dict[str, Any]] = []

            # Safety check - prevent returning too many items
            total_found = len(paths)
            if total_found > self.max_items:
                suggestions = []
                if pattern == "*":
                    suggestions.append("Use a more specific pattern like '*.py', '*.txt', etc.")
                if max_depth > 1:
                    suggestions.append(f"Reduce max_depth from {max_depth} to 1 or 2.")
                if not exclude:
                    suggestions.append("Use exclude parameter to filter out unwanted directories.")

                suggestion_text = " ".join(suggestions) if suggestions else ""
                error_msg = f"Error: Too many items ({total_found:,}). {suggestion_text}"
                return ToolResult(
                    content=error_msg,
                    metadata={"count": total_found, "truncated": True},
                )

            for file_path, file_info in paths.items():  # pyright: ignore[reportAttributeAccessIssue]
                rel_path = os.path.relpath(str(file_path), path)

                # Skip excluded patterns
                if exclude and any(fnmatch(rel_path, pat) for pat in exclude):
                    continue

                # Use type from glob detail info, falling back to isdir only if needed
                is_dir = await is_directory(fs, file_path, entry_type=file_info.get("type"))  # pyright: ignore[reportArgumentType]

                item_info = {
                    "name": Path(file_path).name,  # pyright: ignore[reportArgumentType]
                    "path": file_path,
                    "relative_path": rel_path,
                    "size": file_info.get("size", 0),
                    "type": "directory" if is_dir else "file",
                }
                if "mtime" in file_info:
                    item_info["modified"] = file_info["mtime"]

                if is_dir:
                    dirs.append(item_info)
                else:
                    files.append(item_info)

            await ctx.events.file_operation("list", path=path, success=True)
            result = format_directory_listing(path, dirs, files, pattern)
            # Emit formatted content for UI display
            from agentpool.agents.events import TextContentItem

            await ctx.events.tool_call_progress(
                title=f"Listed: {path}",
                items=[TextContentItem(text=result)],
                replace_content=True,
            )
        except (OSError, ValueError, FileNotFoundError) as e:
            await ctx.events.file_operation("list", path=path, success=False, error=str(e))
            error_msg = (
                f"Error: Could not list directory: {path}. Ensure path is absolute and exists."
            )
            return ToolResult(
                content=error_msg,
                metadata={"count": 0, "truncated": False},
            )
        else:
            total_items = len(files) + len(dirs)
            # Check if we hit the limit (truncation)
            was_truncated = total_items >= self.max_items
            return ToolResult(
                content=result,
                metadata={"count": total_items, "truncated": was_truncated},
            )



"""Claude Code Skill following the Agent Skills Spec."""

from __future__ import annotations

import html
from typing import Annotated, Any
import unicodedata

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from upathtools import UPath


MAX_SKILL_NAME_LENGTH = 64
MAX_DESCRIPTION_LENGTH = 1024
MAX_COMPATIBILITY_LENGTH = 500


class Skill(BaseModel):
    """A Claude Code Skill with metadata and lazy-loaded instructions.

    Follows the Agent Skills Spec for field naming and validation rules.
    Frontmatter fields use ``extra="forbid"`` so unknown YAML keys are rejected.
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    name: Annotated[str, Field(max_length=MAX_SKILL_NAME_LENGTH)]
    description: Annotated[str, Field(min_length=1, max_length=MAX_DESCRIPTION_LENGTH)]
    skill_path: UPath
    license: str | None = None
    compatibility: Annotated[str | None, Field(max_length=MAX_COMPATIBILITY_LENGTH)] = None
    allowed_tools: str | None = Field(default=None, alias="allowed-tools")
    metadata: dict[str, str] = Field(default_factory=dict)
    instructions: str | None = Field(default=None, exclude=True)

    @field_validator("name")
    @classmethod
    def _validate_name(cls, v: str) -> str:
        name = unicodedata.normalize("NFKC", v.strip())
        if not name:
            raise ValueError("Skill name must be non-empty")
        if name != name.lower():
            raise ValueError(f"Skill name {name!r} must be lowercase")
        if name.startswith("-") or name.endswith("-"):
            raise ValueError("Skill name cannot start or end with a hyphen")
        if "--" in name:
            raise ValueError("Skill name cannot contain consecutive hyphens")
        if not all(c.isalnum() or c == "-" for c in name):
            msg = (
                f"Skill name {name!r} contains invalid characters. "
                "Only letters, digits, and hyphens are allowed."
            )
            raise ValueError(msg)
        return name

    @model_validator(mode="before")
    @classmethod
    def _normalize_metadata(cls, data: Any) -> Any:
        if isinstance(data, dict) and "metadata" in data:
            meta = data["metadata"]
            if isinstance(meta, dict):
                data["metadata"] = {str(k): str(v) for k, v in meta.items()}
        return data

    def load_instructions(self) -> str:
        """Lazy-load full instructions from SKILL.md."""
        if self.instructions is None:
            skill_file = self.skill_path / "SKILL.md"
            if skill_file.exists():
                content = skill_file.read_text(encoding="utf-8")
                parts = content.split("---", 2)
                self.instructions = parts[2].strip() if len(parts) >= 3 else ""  # noqa: PLR2004
            else:
                self.instructions = ""
        return self.instructions

    @classmethod
    def from_skill_dir(cls, skill_dir: UPath) -> Skill:
        """Parse a SKILL.md file from a directory and return a validated Skill.

        Args:
            skill_dir: Path to the skill directory containing SKILL.md.

        Returns:
            Validated Skill instance.

        Raises:
            FileNotFoundError: If SKILL.md is not found.
            ValueError: If frontmatter is missing or invalid YAML.
            ValidationError: If fields fail Pydantic validation.
        """
        skill_file = find_skill_md(skill_dir)
        if skill_file is None:
            msg = f"SKILL.md not found in {skill_dir}"
            raise FileNotFoundError(msg)

        content = skill_file.read_text("utf-8")
        metadata, _body = parse_frontmatter(content)
        return cls(skill_path=skill_dir, **metadata)


def find_skill_md(skill_dir: UPath) -> UPath | None:
    """Find the SKILL.md file in a skill directory.

    Prefers SKILL.md (uppercase) but accepts skill.md (lowercase).

    Args:
        skill_dir: Path to the skill directory.

    Returns:
        Path to the SKILL.md file, or None if not found.
    """
    for name in ("SKILL.md", "skill.md"):
        path = skill_dir / name
        if path.exists():
            return path
    return None


def parse_frontmatter(content: str) -> tuple[dict[str, Any], str]:
    """Parse YAML frontmatter from SKILL.md content.

    Args:
        content: Raw content of a SKILL.md file.

    Returns:
        Tuple of (metadata dict, markdown body).

    Raises:
        ValueError: If frontmatter is missing or invalid.
    """
    import yamling

    if not content.startswith("---"):
        raise ValueError("SKILL.md must start with YAML frontmatter (---)")

    parts = content.split("---", 2)
    if len(parts) < 3:  # noqa: PLR2004
        raise ValueError("SKILL.md frontmatter not properly closed with ---")
    try:
        metadata = yamling.load_yaml(parts[1])
    except yamling.YAMLError as e:
        raise ValueError(f"Invalid YAML in frontmatter: {e}") from e

    if not isinstance(metadata, dict):
        raise TypeError("SKILL.md frontmatter must be a YAML mapping")

    return metadata, parts[2].strip()


def to_prompt(skills: list[Skill]) -> str:
    """Generate the ``<available_skills>`` XML block for agent system prompts.

    This XML format is what Anthropic uses and recommends for Claude models.

    Args:
        skills: List of skills to include.

    Returns:
        XML string with ``<available_skills>`` block.
    """
    if not skills:
        return "<available_skills>\n</available_skills>"

    lines = ["<available_skills>"]
    for skill in skills:
        lines.append("<skill>")
        lines.append(f"<name>{html.escape(skill.name)}</name>")
        lines.append(f"<description>{html.escape(skill.description)}</description>")
        skill_md = find_skill_md(skill.skill_path)
        if skill_md is not None:
            lines.append(f"<location>{skill_md}</location>")
        lines.append("</skill>")
    lines.append("</available_skills>")
    return "\n".join(lines)



"""Syntax highlighting utilities for file content display."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path

DOT_FILE_TO_LANGUAGE = {
    ".bashrc": "bash",
    ".bash_profile": "bash",
    ".bash_logout": "bash",
    ".zshrc": "bash",
    ".zsh_profile": "bash",
    ".profile": "bash",
    ".vimrc": "vim",
    ".gvimrc": "vim",
    ".tmux.conf": "tmux",
    ".gitignore": "gitignore",
    ".gitconfig": "gitconfig",
    ".editorconfig": "editorconfig",
    ".eslintrc": "json",
    ".prettierrc": "json",
    ".babelrc": "json",
}

EXT_TO_LANGUAGE = {
    # Python
    ".py": "python",
    ".pyx": "python",
    ".pyi": "python",
    # JavaScript/TypeScript
    ".js": "javascript",
    ".mjs": "javascript",
    ".jsx": "javascript",
    ".ts": "typescript",
    ".tsx": "typescript",
    ".mts": "typescript",
    # JVM Languages
    ".java": "java",
    ".kt": "kotlin",
    ".scala": "scala",
    ".groovy": "groovy",
    # C/C++
    ".c": "c",
    ".h": "c",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".hpp": "cpp",
    ".hxx": "cpp",
    # Rust
    ".rs": "rust",
    # Go
    ".go": "go",
    # Web
    ".html": "html",
    ".htm": "html",
    ".xml": "xml",
    ".xhtml": "html",
    ".css": "css",
    ".scss": "scss",
    ".sass": "sass",
    ".less": "less",
    # Data formats
    ".json": "json",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".toml": "toml",
    ".csv": "csv",
    # Shell/Scripts
    ".sh": "bash",
    ".bash": "bash",
    ".zsh": "bash",
    ".fish": "fish",
    ".ps1": "powershell",
    ".psm1": "powershell",
    ".bat": "batch",
    ".cmd": "batch",
    # Other languages
    ".rb": "ruby",
    ".php": "php",
    ".sql": "sql",
    ".r": "r",
    ".swift": "swift",
    ".dart": "dart",
    ".lua": "lua",
    ".pl": "perl",
    ".pm": "perl",
    ".vim": "vim",
    ".tex": "latex",
    ".cls": "latex",
    ".sty": "latex",
    # Config/Other
    ".ini": "ini",
    ".cfg": "ini",
    ".conf": "conf",
    ".log": "log",
    ".md": "markdown",
    ".markdown": "markdown",
    ".mdown": "markdown",
    ".mkd": "markdown",
    ".rst": "rst",
    ".txt": "text",
    # Docker
    ".dockerfile": "dockerfile",
    # Makefiles
    ".mk": "makefile",
    # Nginx
    ".nginx": "nginx",
    # Apache
    ".htaccess": "apache",
    # Git
    ".gitignore": "gitignore",
    ".gitconfig": "gitconfig",
}


def get_language_from_path(path: str | Path) -> str:  # noqa: PLR0911
    """Get syntax highlighting language identifier from file path.

    Args:
        path: File path to analyze

    Returns:
        Language identifier for syntax highlighting, or empty string if unknown

    Examples:
        >>> get_language_from_path("src/main.py")
        'python'
        >>> get_language_from_path("config.json")
        'json'
        >>> get_language_from_path("Dockerfile")
        'dockerfile'
        >>> get_language_from_path("unknown.xyz")
        ''
    """
    path_str = str(path)
    _, ext = os.path.splitext(path_str.lower())  # noqa: PTH122
    # Special cases for files without extensions or specific names
    filename = os.path.basename(path_str.lower())  # noqa: PTH119
    # Docker files
    if filename in ("dockerfile", "dockerfile.dev", "dockerfile.prod"):
        return "dockerfile"
    # Build files
    if filename in ("makefile", "rakefile", "justfile"):
        return "makefile"
    if filename in (".rules", ".cursorrules"):
        return "markdown"
    # License files
    if filename in ("license", "licence", "copying", "contributing"):
        return "text"
    if filename == "yarn.lock":
        return "yaml"
    # Config files starting with dot
    if filename in DOT_FILE_TO_LANGUAGE:
        return DOT_FILE_TO_LANGUAGE[filename]
    # README files
    if filename.startswith("readme") and "." not in filename:
        return "text"
    # Package files
    return EXT_TO_LANGUAGE.get(ext, "")


def format_code_block(content: str, language: str = "") -> str:
    r"""Format content as a markdown code block with optional language.

    Args:
        content: The code/text content
        language: Optional language identifier for syntax highlighting

    Returns:
        Formatted markdown code block

    Examples:
        >>> format_code_block('print("hello")', 'python')
        '```python\\nprint("hello")\\n```'
        >>> format_code_block('some text')
        '```\\nsome text\\n```'
    """
    return f"```{language}\n{content}\n```"


def format_file_content(content: str, file_path: str | Path) -> str:
    r"""Format file content as a syntax-highlighted code block.

    Args:
        content: The file content
        file_path: Path to the file (used to determine language)

    Returns:
        Formatted markdown code block with appropriate syntax highlighting

    Examples:
        >>> format_file_content('def hello(): pass', 'main.py')
        '```python\\ndef hello(): pass\\n```'
    """
    language = get_language_from_path(file_path)
    return format_code_block(content, language)


def format_zed_code_block(
    content: str,
    file_path: str | Path,
    start_line: int | None = None,
    end_line: int | None = None,
) -> str:
    r"""Format content as a Zed-compatible code block with path and optional line numbers.

    Args:
        content: The code/text content
        file_path: Path to the file (used in code block header)
        start_line: Optional starting line number (1-based)
        end_line: Optional ending line number (1-based)

    Returns:
        Formatted Zed-compatible code block

    Examples:
        >>> format_zed_code_block('def hello(): pass', 'src/main.py')
        '```src/main.py\\ndef hello(): pass\\n```'
        >>> format_zed_code_block('def hello(): pass', 'src/main.py', 10)
        '```src/main.py#L10\\ndef hello(): pass\\n```'
        >>> format_zed_code_block('def hello(): pass', 'src/main.py', 10, 15)
        '```src/main.py#L10-15\\ndef hello(): pass\\n```'
    """
    path_str = str(file_path)

    # Build line number suffix if provided
    line_suffix = ""
    if start_line is not None:
        if end_line is not None and end_line != start_line:
            line_suffix = f"#L{start_line}-{end_line}"
        else:
            line_suffix = f"#L{start_line}"

    return f"```{path_str}{line_suffix}\n{content}\n```"



"""Mode categories for ACPAgent."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from agentpool.agents.exceptions import UnknownModeError
from agentpool.agents.modes import ConfigOptionChanged, ModeCategoryProtocol, ModeInfo


if TYPE_CHECKING:
    from acp.schema import SessionConfigOption, SessionModelState, SessionModeState
    from agentpool.agents.acp_agent.acp_agent import ACPAgent


# =============================================================================
# Mode category implementations
# =============================================================================


@dataclass
class ACPModeCategory(ModeCategoryProtocol["ACPAgent"]):
    """Mode category for ACP - handles permissions/approval policy.

    Automatically uses config_options API if available, falls back to
    set_session_mode for older servers.
    """

    available_modes: list[ModeInfo]

    id: ClassVar[str] = "mode"
    name: ClassVar[str] = "Mode"
    category: ClassVar[str] = "mode"

    def get_current(self, agent: ACPAgent) -> str:
        """Get current mode from ACP state."""
        # Try config_options first
        if agent._state and agent._state.config_options:
            for opt in agent._state.config_options:
                if opt.id == "mode":
                    return opt.current_value
        # Fall back to legacy modes state
        if agent._state and agent._state.modes:
            return agent._state.modes.current_mode_id
        return ""

    async def apply(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply mode change - uses config_options if available, else legacy API."""
        valid_ids = {m.id for m in self.available_modes}
        if mode_id not in valid_ids:
            raise UnknownModeError(mode_id, list(valid_ids))

        if not agent._connection or not agent._sdk_session_id or not agent._state:
            raise RuntimeError("Not connected to ACP server")

        # Try config_options API first
        if agent._state.config_options:
            await self._apply_via_config_options(agent, mode_id)
        else:
            await self._apply_via_legacy(agent, mode_id)

        await agent.update_state(config_id=self.id, value_id=mode_id)

    async def _apply_via_config_options(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply using new config_options API."""
        from acp.schema import SetSessionConfigOptionRequest

        assert agent._connection is not None
        assert agent._state is not None
        assert agent._sdk_session_id is not None

        config_request = SetSessionConfigOptionRequest(
            session_id=agent._sdk_session_id,
            config_id="mode",
            value=mode_id,
        )
        response = await agent._connection.set_session_config_option(config_request)

        if response.config_options:
            agent._state.config_options = list(response.config_options)

        agent.log.info("Mode changed via config_options", mode_id=mode_id)

    async def _apply_via_legacy(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply using legacy set_session_mode API."""
        from acp.schema import SetSessionModeRequest

        assert agent._connection is not None
        assert agent._state is not None
        assert agent._sdk_session_id is not None

        mode_request = SetSessionModeRequest(session_id=agent._sdk_session_id, mode_id=mode_id)
        await agent._connection.set_session_mode(mode_request)

        if agent._state.modes:
            agent._state.modes.current_mode_id = mode_id

        agent.log.info("Mode changed via legacy API", mode_id=mode_id)

    @classmethod
    def from_state(
        cls,
        config_options: list[SessionConfigOption] | None,
        modes_state: SessionModeState | None,
    ) -> ACPModeCategory | None:
        """Create from ACP state - prefers config_options, falls back to modes_state.

        Args:
            config_options: SessionConfigOption list (new API)
            modes_state: SessionModeState (legacy API)

        Returns:
            ACPModeCategory instance, or None if no mode info available
        """
        from acp.schema import SessionConfigSelectGroup

        # Try config_options first
        if config_options:
            for config_opt in config_options:
                if config_opt.id != "mode":
                    continue
                mode_infos: list[ModeInfo] = []
                for opt_item in config_opt.options:
                    if isinstance(opt_item, SessionConfigSelectGroup):
                        mode_infos.extend(
                            ModeInfo(
                                id=sub_opt.value,
                                name=sub_opt.name,
                                description=sub_opt.description or "",
                                category_id="mode",
                            )
                            for sub_opt in opt_item.options
                        )
                    else:
                        mode_infos.append(
                            ModeInfo(
                                id=opt_item.value,
                                name=opt_item.name,
                                description=opt_item.description or "",
                                category_id="mode",
                            )
                        )
                return cls(available_modes=mode_infos)

        # Fall back to legacy modes state
        if modes_state:
            modes = [
                ModeInfo(
                    id=m.id,
                    name=m.name,
                    description=m.description or "",
                    category_id="mode",
                )
                for m in modes_state.available_modes
            ]
            return cls(available_modes=modes)

        return None


@dataclass
class ACPModelCategory(ModeCategoryProtocol["ACPAgent"]):
    """Model category for ACP - handles model selection.

    Automatically uses config_options API if available, falls back to
    set_session_model for older servers.
    """

    available_modes: list[ModeInfo]

    id: ClassVar[str] = "model"
    name: ClassVar[str] = "Model"
    category: ClassVar[str] = "model"

    def get_current(self, agent: ACPAgent) -> str:
        """Get current model from ACP state."""
        # Try config_options first
        if agent._state and agent._state.config_options:
            for opt in agent._state.config_options:
                if opt.id == "model":
                    return opt.current_value
        # Fall back to legacy models state
        if agent._state and agent._state.models:
            return agent._state.models.current_model_id
        return ""

    async def apply(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply model change - uses config_options if available, else legacy API."""
        valid_ids = {m.id for m in self.available_modes}
        if mode_id not in valid_ids:
            raise UnknownModeError(mode_id, list(valid_ids))

        if not agent._connection or not agent._sdk_session_id or not agent._state:
            raise RuntimeError("Not connected to ACP server")

        # Try config_options API first
        if agent._state.config_options:
            await self._apply_via_config_options(agent, mode_id)
        else:
            await self._apply_via_legacy(agent, mode_id)
        await agent.update_state(config_id=self.id, value_id=mode_id)

    async def _apply_via_config_options(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply using new config_options API."""
        from acp.schema import SetSessionConfigOptionRequest

        assert agent._connection is not None
        assert agent._state is not None
        assert agent._sdk_session_id is not None

        config_request = SetSessionConfigOptionRequest(
            session_id=agent._sdk_session_id,
            config_id="model",
            value=mode_id,
        )
        response = await agent._connection.set_session_config_option(config_request)

        if response.config_options:
            agent._state.config_options = list(response.config_options)

        agent.log.info("Model changed via config_options", model_id=mode_id)

    async def _apply_via_legacy(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply using legacy set_session_model API."""
        from acp.schema import SetSessionModelRequest

        assert agent._connection is not None
        assert agent._state is not None
        assert agent._sdk_session_id is not None

        request = SetSessionModelRequest(session_id=agent._sdk_session_id, model_id=mode_id)
        if await agent._connection.set_session_model(request):
            agent._state.current_model_id = mode_id
            agent.log.info("Model changed via legacy API", model_id=mode_id)
        else:
            msg = (
                "Remote ACP agent does not support model changes. "
                "set_session_model returned no response."
            )
            raise RuntimeError(msg)

    @classmethod
    def from_state(
        cls,
        config_options: list[SessionConfigOption] | None,
        models_state: SessionModelState | None,
    ) -> ACPModelCategory | None:
        """Create from ACP state - prefers config_options, falls back to models_state.

        Args:
            config_options: SessionConfigOption list (new API)
            models_state: SessionModelState (legacy API)

        Returns:
            ACPModelCategory instance, or None if no model info available
        """
        from acp.schema import SessionConfigSelectGroup

        # Try config_options first
        if config_options:
            for config_opt in config_options:
                if config_opt.id != "model":
                    continue
                mode_infos: list[ModeInfo] = []
                if isinstance(config_opt.options, list):
                    for opt_item in config_opt.options:
                        if isinstance(opt_item, SessionConfigSelectGroup):
                            mode_infos.extend(
                                ModeInfo(
                                    id=sub_opt.value,
                                    name=sub_opt.name,
                                    description=sub_opt.description or "",
                                    category_id="model",
                                )
                                for sub_opt in opt_item.options
                            )
                        else:
                            mode_infos.append(
                                ModeInfo(
                                    id=opt_item.value,
                                    name=opt_item.name,
                                    description=opt_item.description or "",
                                    category_id="model",
                                )
                            )
                return cls(available_modes=mode_infos)

        # Fall back to legacy models state
        if models_state:
            models = [
                ModeInfo(
                    id=m.model_id,
                    name=m.name,
                    description=m.description or "",
                    category_id="model",
                )
                for m in models_state.available_models
            ]
            return cls(available_modes=models)

        return None


@dataclass
class ACPGenericCategory(ModeCategoryProtocol["ACPAgent"]):
    """Generic category for ACP - handles any config_option that's not mode/model.

    For things like thought_level or custom categories that only exist in
    the config_options API.
    """

    id: str
    name: str
    available_modes: list[ModeInfo]
    category: str | None = None

    def get_current(self, agent: ACPAgent) -> str:
        """Get current value from ACP state."""
        if agent._state and (opts := agent._state.config_options):
            return next((i.current_value for i in opts if i.id == self.id), "")
        return ""

    async def apply(self, agent: ACPAgent, mode_id: str) -> None:
        """Apply config option change via config_options API."""
        from acp.schema import SetSessionConfigOptionRequest

        valid_ids = {m.id for m in self.available_modes}
        if mode_id not in valid_ids:
            raise UnknownModeError(mode_id, list(valid_ids))

        if not agent._connection or not agent._sdk_session_id or not agent._state:
            raise RuntimeError("Not connected to ACP server")

        if not agent._state.config_options:
            raise RuntimeError(f"Server does not support config_options, cannot set {self.id!r}")

        config_request = SetSessionConfigOptionRequest(
            session_id=agent._sdk_session_id,
            config_id=self.id,
            value=mode_id,
        )
        response = await agent._connection.set_session_config_option(config_request)

        if response.config_options:
            agent._state.config_options = list(response.config_options)

        agent.log.info("Config option changed", config_id=self.id, value=mode_id)
        change = ConfigOptionChanged(config_id=self.id, value_id=mode_id)
        await agent.state_updated.emit(change)

    @classmethod
    def from_config_options(
        cls,
        config_options: list[SessionConfigOption],
    ) -> list[ACPGenericCategory]:
        """Create categories for non-mode/model config options.

        Args:
            config_options: SessionConfigOption list

        Returns:
            List of ACPGenericCategory for options that aren't mode/model
        """
        from acp.schema import SessionConfigSelectGroup

        categories: list[ACPGenericCategory] = []

        for config_opt in config_options:
            # Skip mode and model - they have dedicated categories
            if config_opt.id in ("mode", "model"):
                continue

            mode_infos: list[ModeInfo] = []
            if isinstance(config_opt.options, list):
                for opt_item in config_opt.options:
                    if isinstance(opt_item, SessionConfigSelectGroup):
                        mode_infos.extend(
                            ModeInfo(
                                id=sub_opt.value,
                                name=sub_opt.name,
                                description=sub_opt.description or "",
                                category_id=config_opt.id,
                            )
                            for sub_opt in opt_item.options
                        )
                    else:
                        mode_infos.append(
                            ModeInfo(
                                id=opt_item.value,
                                name=opt_item.name,
                                description=opt_item.description or "",
                                category_id=config_opt.id,
                            )
                        )

            categories.append(
                cls(
                    id=config_opt.id,
                    name=config_opt.name,
                    available_modes=mode_infos,
                    category=config_opt.category or "other",
                )
            )

        return categories
